# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import inspect
import logging
import sys
import textwrap

import pyomo.core.expr as EXPR
import pyomo.core.base as BASE
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice
from pyomo.core.base.initializer import Initializer
from pyomo.core.base.component import Component, ActiveComponent, ComponentData
from pyomo.core.base.config import PyomoOptions
from pyomo.core.base.enums import SortComponents
from pyomo.core.base.global_set import UnindexedComponent_set
from pyomo.core.expr.numeric_expr import _ndarray
from pyomo.core.pyomoobject import PyomoObject
from pyomo.common import DeveloperError
from pyomo.common.autoslots import fast_deepcopy
from pyomo.common.collections import ComponentSet
from pyomo.common.deprecation import deprecated, deprecation_warning
from pyomo.common.errors import TemplateExpressionError
from pyomo.common.modeling import NOTSET
from pyomo.common.numeric_types import native_types
from pyomo.common.sorting import sorted_robust

from collections.abc import Sequence

logger = logging.getLogger('pyomo.core')

sequence_types = {tuple, list}
slicer_types = {slice, Ellipsis.__class__, IndexedComponent_slice}


def normalize_index(x):
    """Normalize a component index.

    This flattens nested sequences into a single tuple.  There is a
    "global" flag (normalize_index.flatten) that will turn off index
    flattening across Pyomo.

    Scalar values will be returned unchanged.  Tuples with a single
    value will be unpacked and returned as a single value.

    Returns
    -------
    scalar or tuple

    """
    if x.__class__ in native_types:
        return x
    elif x.__class__ in sequence_types:
        # Note that casting a tuple to a tuple is cheap (no copy, no
        # new object)
        x = tuple(x)
    else:
        # Note: new Sequence types will be caught below and added to the
        # sequence_types set
        x = (x,)

    x_len = len(x)
    i = 0
    while i < x_len:
        _xi_class = x[i].__class__
        if _xi_class in native_types:
            i += 1
        elif _xi_class in sequence_types:
            x_len += len(x[i]) - 1
            # Note that casting a tuple to a tuple is cheap (no copy, no
            # new object)
            x = x[:i] + tuple(x[i]) + x[i + 1 :]
        elif issubclass(_xi_class, Sequence):
            if issubclass(_xi_class, str):
                # This is very difficult to get to: it would require a
                # user creating a custom derived string type
                native_types.add(_xi_class)
                i += 1
            else:
                sequence_types.add(_xi_class)
                x_len += len(x[i]) - 1
                x = x[:i] + tuple(x[i]) + x[i + 1 :]
        else:
            i += 1

    if x_len == 1:
        return x[0]
    return x


# Pyomo will normalize indices by default
normalize_index.flatten = True


class _NotFound:
    pass


class _NotSpecified:
    pass


#
# Get the fully-qualified name for this index.  If there isn't anything
# in the _data dict (and there shouldn't be), then add something, get
# the name, and remove it.  This allows us to get the name of something
# that we haven't added yet without changing the state of the constraint
# object.
#
def _get_indexed_component_data_name(component, index):
    """Returns the fully-qualified component name for an unconstructed index.

    The ComponentData.name property assumes that the ComponentData has
    already been assigned to the owning Component.  This is a problem
    during the process of constructing a ComponentData instance, as we
    may need to throw an exception before the ComponentData is added to
    the owning component.  In those cases, we can use this function to
    generate the fully-qualified name without (permanently) adding the
    object to the Component.

    """
    if not component.is_indexed():
        return component.name
    elif index in component._data:
        ans = component._data[index].name
    else:
        for i in range(5):
            try:
                component._data[index] = component._ComponentDataClass(
                    *((None,) * i), component=component
                )
                i = None
                break
            except:
                pass
        if i is not None:
            # None of the generic positional arguments worked; raise an
            # exception
            component._data[index] = component._ComponentDataClass(component=component)
        try:
            ans = component._data[index].name
        except:
            ans = component.name + '[{unknown index}]'
        finally:
            del component._data[index]
    return ans


_rule_returned_none_error = """%s '%s': rule returned None.

%s rules must return either a valid expression, numeric value, or
%s.Skip.  The most common cause of this error is forgetting to
include the "return" statement at the end of your rule.
"""


def rule_result_substituter(result_map, map_types):
    _map = result_map
    if map_types is None:
        _map_types = set(type(key) for key in result_map)
    else:
        _map_types = map_types

    def rule_result_substituter_impl(rule, *args, **kwargs):
        if rule.__class__ in _map_types:
            #
            # The argument is a trivial type and will be mapped
            #
            value = rule
        elif isinstance(rule, PyomoObject):
            #
            # The argument is a Pyomo component.  This can happen when
            # the rule isn't a rule at all, but instead the decorator
            # was used as a function to wrap an inline definition (not
            # something I think we should support, but exists in some
            # [old] examples).
            #
            return rule
        else:
            #
            # Otherwise, the argument is a functor, so call it to
            # generate the rule result.
            #
            value = rule(*args, **kwargs)
        #
        # Map the returned value:
        #
        if value.__class__ in _map_types and value in _map:
            return _map[value]
        return value

    return rule_result_substituter_impl


_map_rule_funcdef = """def wrapper_function%s:
    args, varargs, kwds, local_env = inspect.getargvalues(
        inspect.currentframe())
    args = tuple(local_env[_] for _ in args) + (varargs or ())
    return wrapping_fcn(rule, *args, **(kwds or {}))
"""


def rule_wrapper(rule, wrapping_fcn, positional_arg_map=None, map_types=None):
    """Wrap a rule with another function

    This utility method provides a way to wrap a function (rule) with
    another function while preserving the original function signature.
    This is important for rules, as the :py:func:`Initializer`
    argument processor relies on knowing the number of positional
    arguments.

    Parameters
    ----------
    rule: function
        The original rule being wrapped
    wrapping_fcn: function or Dict
        The wrapping function.  The `wrapping_fcn` will be called with
        ``(rule, *args, **kwargs)``.  For convenience, if a `dict` is
        passed as the `wrapping_fcn`, then the result of
        :py:func:`rule_result_substituter(wrapping_fcn)` is used as the
        wrapping function.
    positional_arg_map: iterable[int]
        An iterable of indices of rule positional arguments to expose in
        the wrapped function signature.  For example,
        `positional_arg_map=(2, 0)` and `rule=fcn(a, b, c)` would produce a
        wrapped function with a signature `wrapper_function(c, a)`

    """
    if isinstance(wrapping_fcn, dict):
        wrapping_fcn = rule_result_substituter(wrapping_fcn, map_types)
        if not inspect.isfunction(rule):
            return wrapping_fcn(rule)
    # Because some of our processing of initializer functions relies on
    # knowing the number of positional arguments, we will go to extra
    # effort here to preserve the original function signature.
    rule_sig = inspect.signature(rule)
    if positional_arg_map is not None:
        param = list(rule_sig.parameters.values())
        rule_sig = rule_sig.replace(parameters=(param[i] for i in positional_arg_map))
    _funcdef = _map_rule_funcdef % (str(rule_sig),)
    # Create the wrapper in a temporary environment that mimics this
    # function's environment.
    _env = dict(globals())
    _env.update(locals())
    exec(_funcdef, _env)
    return _env['wrapper_function']


class IndexedComponent(Component):
    """This is the base class for all indexed modeling components.
    This class stores a dictionary, self._data, that maps indices
    to component data objects.  The object self._index_set defines valid
    keys for this dictionary, and the dictionary keys may be a
    strict subset.

    The standard access and iteration methods iterate over the the
    keys of self._data.  This class supports a concept of a default
    component data value.  When enabled, the default does not
    change the access and iteration methods.

    IndexedComponent may be given a set over which indexing is restricted.
    Alternatively, IndexedComponent may be indexed over Any
    (pyomo.core.base.set_types.Any), in which case the IndexedComponent
    behaves like a dictionary - any hashable object can be used as a key
    and keys can be added on the fly.

    Constructor arguments:
        ctype       The class type for the derived subclass
        doc         A text string describing this component

    Private class attributes:

        _data:  A dictionary from the index set to component data objects

        _index_set:  The set of valid indices

        _anonymous_sets: A ComponentSet of "anonymous" sets used by this
            component.  Anonymous sets are Set / SetOperator / RangeSet
            that compose attributes like _index_set, but are not
            themselves explicitly assigned (and named) on any Block

    """

    class Skip:
        pass

    #
    # If an index is supplied for which there is not a _data entry
    # (specifically, in a get call), then this flag determines whether
    # a check is performed to see if the input index is in the
    # index set _index_set. This is extremely expensive, and so this flag
    # is provided to disable that feature globally.
    #
    _DEFAULT_INDEX_CHECKING_ENABLED = True

    def __init__(self, *args, **kwds):
        #
        kwds.pop('noruleinit', None)
        Component.__init__(self, **kwds)
        #
        self._data = {}
        #
        if len(args) == 0 or (args[0] is UnindexedComponent_set and len(args) == 1):
            #
            # If no indexing sets are provided, generate a dummy index
            #
            self._index_set = UnindexedComponent_set
            self._anonymous_sets = None
        elif len(args) == 1:
            #
            # If a single indexing set is provided, just process it.
            #
            self._index_set, self._anonymous_sets = BASE.set.process_setarg(args[0])
        else:
            #
            # If multiple indexing sets are provided, process them all,
            # and store the cross-product of these sets.
            #
            # Example: Pyomo allows things like "Param([1,2,3],
            # range(100), initialize=0)".  This needs to create *3*
            # sets: two SetOf components and then the SetProduct.  As
            # the user declined to name any of these sets, we will not
            # make up names and instead store them on the model as
            # "anonymous components"
            #
            self._index_set = BASE.set.SetProduct(*args)
            self._anonymous_sets = ComponentSet((self._index_set,))
            if self._index_set._anonymous_sets is not None:
                self._anonymous_sets.update(self._index_set._anonymous_sets)

    def _create_objects_for_deepcopy(self, memo, component_list):
        _new = self.__class__.__new__(self.__class__)
        _ans = memo.setdefault(id(self), _new)
        if _ans is _new:
            component_list.append((self, _new))
            # For indexed components, we will pre-emptively clone all
            # component data objects as well (as those are the objects
            # that will be referenced by things like expressions).  It
            # is important to only clone "normal" ComponentData objects:
            # so we will want to skip this for all scalar components
            # (where the _data points back to self) and references
            # (where the data may be stored outside this block tree and
            # therefore may not be cloned)
            if self.is_indexed() and not self.is_reference():
                # Because we are already checking / updating the memo
                # for the _data dict, we can effectively "deepcopy" it
                # right now (almost for free!)
                _src = self._data
                memo[id(_src)] = _new._data = _src.__class__()
                _setter = _new._data.__setitem__
                for idx, obj in _src.items():
                    _setter(
                        fast_deepcopy(idx, memo),
                        obj._create_objects_for_deepcopy(memo, component_list),
                    )

        return _ans

    def to_dense_data(self):
        """TODO"""
        for idx in self._index_set:
            if idx in self._data:
                continue
            try:
                self._getitem_when_not_present(idx)
            except KeyError:
                # Rule could have returned Skip, which we will silently ignore
                pass

    def clear(self):
        """Clear the data in this component"""
        if self.is_indexed():
            self._data = {}
        else:
            raise DeveloperError(
                "Derived scalar component %s failed to define clear()."
                % (self.__class__.__name__,)
            )

    def index_set(self):
        """Return the index set"""
        return self._index_set

    def is_indexed(self):
        """Return true if this component is indexed"""
        return self._index_set is not UnindexedComponent_set

    def is_reference(self):
        """Return True if this component is a reference, where
        "reference" is interpreted as any component that does not
        own its own data.
        """
        return self._data is not None and type(self._data) is not dict

    def dim(self):
        """Return the dimension of the index"""
        if not self.is_indexed():
            return 0
        return self._index_set.dimen

    def __len__(self):
        """
        Return the number of component data objects stored by this
        component.
        """
        return len(self._data)

    def __contains__(self, idx):
        """Return true if the index is in the dictionary"""
        return idx in self._data

    # The default implementation is for keys() and __iter__ to be
    # synonyms.  The logic is implemented in keys() so that
    # keys/values/items continue to work for components that implement
    # other definitions for __iter__ (e.g., Set)
    def __iter__(self):
        """Return an iterator of the component data keys"""
        return self.keys()

    def keys(self, sort=SortComponents.UNSORTED, ordered=NOTSET):
        """Return an iterator over the component data keys

        This method sets the ordering of component data objects within
        this IndexedComponent container.  For consistency,
        :py:meth:`__init__()`, :py:meth:`values`, and :py:meth:`items`
        all leverage this method to ensure consistent ordering.

        Parameters
        ----------
        sort: bool or SortComponents
            Iterate over the declared component keys in a specified
            sorted order.  See :py:class:`SortComponents` for valid
            options and descriptions.

        ordered: bool
            DEPRECATED: Please use `sort=SortComponents.ORDERED_INDICES`.
            If True, then the keys are returned in a deterministic order
            (using the underlying set's `ordered_iter()`).

        """
        sort = SortComponents(sort)
        if ordered is not NOTSET:
            deprecation_warning(
                f"keys(ordered={ordered}) is deprecated.  "
                "Please use `sort=SortComponents.ORDERED_INDICES`",
                version='6.6.0',
            )
            if ordered:
                sort = sort | SortComponents.ORDERED_INDICES
        if not self._index_set.isfinite():
            #
            # If the index set is virtual (e.g., Any) then return the
            # data iterator.  Note that since we cannot check the length
            # of the underlying Set, there should be no warning if the
            # user iterates over the set when the _data dict is empty.
            #
            if (
                SortComponents.SORTED_INDICES in sort
                or SortComponents.ORDERED_INDICES in sort
            ):
                return iter(sorted_robust(self._data))
            else:
                return self._data.__iter__()

        if SortComponents.SORTED_INDICES in sort:
            ans = self._index_set.sorted_iter()
        elif SortComponents.ORDERED_INDICES in sort:
            ans = self._index_set.ordered_iter()
        else:
            ans = iter(self._index_set)

        if self._data.__class__ is not dict:
            # We currently only need to worry about sparse data
            # structures when the underlying _data is a dict.  Avoiding
            # the len() and filter() below is especially important for
            # References (where both can be expensive linear-time
            # operations)
            pass
        elif len(self) == len(self._index_set):
            #
            # If the data is dense then return the index iterator.
            #
            pass
        elif not self._data and self._index_set and PyomoOptions.paranoia_level:
            logger.warning("""Iterating over a Component (%s)
defined by a non-empty concrete set before any data objects have
actually been added to the Component.  The iterator will be empty.
This is usually caused by Concrete models where you declare the
component (e.g., a Var) and apply component-level operations (e.g.,
x.fix(0)) before you use the component members (in something like a
constraint).

You can silence this warning by one of three ways:
    1) Declare the component to be dense with the 'dense=True' option.
       This will cause all data objects to be immediately created and
       added to the Component.
    2) Defer component-level iteration until after the component data
       members have been added (through explicit use).
    3) If you intend to iterate over a component that may be empty, test
       if the component is empty first and avoid iteration in the case
       where it is empty.
""" % (self.name,))
            return iter(self._data)
        elif SortComponents.SORTED_INDICES in sort:
            # We are sorting the indices (and this is a sparse
            # IndexedComponent): we might as well just sort the sparse
            # _data keys instead of iterating over the whole index.
            return iter(sorted_robust(self._data))
        else:
            #
            # Test each element of a sparse _data with an ordered
            # index set in order.  This is potentially *slow*: if
            # the component is in fact very sparse, we could be
            # iterating over a huge (dense) index in order to sort a
            # small number of indices.  However, this provides a
            # consistent ordering that the user expects.
            #
            ans = filter(self._data.__contains__, ans)
        return ans

    def values(self, sort=SortComponents.UNSORTED, ordered=NOTSET):
        """Return an iterator of the component data objects

        Parameters
        ----------
        sort: bool or SortComponents
            Iterate over the declared component values in a specified
            sorted order.  See :py:class:`SortComponents` for valid
            options and descriptions.

        ordered: bool
            DEPRECATED: Please use `sort=SortComponents.ORDERED_INDICES`.
            If True, then the values are returned in a deterministic order
            (using the underlying set's `ordered_iter()`.
        """
        if ordered is not NOTSET:
            deprecation_warning(
                f"values(ordered={ordered}) is deprecated.  "
                "Please use `sort=SortComponents.ORDERED_INDICES`",
                version='6.6.0',
            )
            if ordered:
                sort = SortComponents(sort) | SortComponents.ORDERED_INDICES
        # Note that looking up the values in a reference may be an
        # expensive operation (linear time).  To avoid making this a
        # quadratic time operation, we will leverage _ReferenceDict's
        # values().  This may fail for references created from mappings
        # or sequences, raising the TypeError
        if self.is_reference():
            try:
                return self._data.values(sort)
            except TypeError:
                pass
        # We would like to look things up directly in _data (as that is
        # fast, since we know that keys() will return valid entries).
        # However, some components (notably Param with a default) have
        # valid keys that do not have an entry in _data.  If we just
        # rely on getitem, then we will hit issues for abstract
        # components.  So we will use a custom getter that tries _data
        # first (both for efficiency and to correctly handle
        # AbstractScalar components), and then falls back on getitem.
        _getdata = self._data.__getitem__

        def getter(s):
            try:
                return _getdata(s)
            except KeyError:
                return self[s]

        return map(getter, self.keys(sort))

    def items(self, sort=SortComponents.UNSORTED, ordered=NOTSET):
        """Return an iterator of (index,data) component data tuples

        Parameters
        ----------
        sort: bool or SortComponents
            Iterate over the declared component items in a specified
            sorted order.  See :py:class:`SortComponents` for valid
            options and descriptions.

        ordered: bool
            DEPRECATED: Please use `sort=SortComponents.ORDERED_INDICES`.
            If True, then the items are returned in a deterministic order
            (using the underlying set's `ordered_iter()`.
        """
        if ordered is not NOTSET:
            deprecation_warning(
                f"items(ordered={ordered}) is deprecated.  "
                "Please use `sort=SortComponents.ORDERED_INDICES`",
                version='6.6.0',
            )
            if ordered:
                sort = SortComponents(sort) | SortComponents.ORDERED_INDICES
        # Note that looking up the values in a reference may be an
        # expensive operation (linear time).  To avoid making this a
        # quadratic time operation, we will try and use _ReferenceDict's
        # items().  This may fail for references created from mappings
        # or sequences, raising the TypeError
        if self.is_reference():
            try:
                return self._data.items(sort)
            except TypeError:
                pass
        # We would like to look things up directly in _data (as that is
        # fast, since we know that keys() will return valid entries).
        # However, some components (notably Param with a default) have
        # valid keys that do not have an entry in _data.  If we just
        # rely on getitem, then we will hit issues in pprint for
        # abstract components.  So we will use a custom getter that
        # tries _data first (both for efficiency and to correctly handle
        # AbstractScalar components), and then falls back on getitem.
        _getdata = self._data.__getitem__

        def getter(s):
            try:
                return s, _getdata(s)
            except KeyError:
                return s, self[s]

        return map(getter, self.keys(sort))

    @deprecated('The iterkeys method is deprecated. Use dict.keys().', version='6.0')
    def iterkeys(self):
        """Return a list of keys in the dictionary"""
        return self.keys()

    @deprecated(
        'The itervalues method is deprecated. Use dict.values().', version='6.0'
    )
    def itervalues(self):
        """Return a list of the component data objects in the dictionary"""
        return self.values()

    @deprecated('The iteritems method is deprecated. Use dict.items().', version='6.0')
    def iteritems(self):
        """Return a list (index,data) tuples from the dictionary"""
        return self.items()

    def __getitem__(self, index) -> ComponentData:
        """
        This method returns the data corresponding to the given index.
        """
        if self._constructed is False:
            self._not_constructed_error(index)

        try:
            return self._data[index]
        except KeyError:
            obj = _NotFound
        except TypeError:
            try:
                index = self._processUnhashableIndex(index)
            except TypeError:
                # This index is really unhashable.  Set a flag so that
                # we can re-raise the original exception (not this one)
                index = TypeError
            if index is TypeError:
                raise
            if index.__class__ is IndexedComponent_slice:
                return index
            # The index could have contained constant but nonhashable
            # objects (e.g., scalar immutable Params).
            # _processUnhashableIndex will evaluate those constants, so
            # if it made any changes to the index, we need to re-check
            # the _data dict for membership.
            try:
                obj = self._data.get(index, _NotFound)
            except TypeError:
                obj = _NotFound

        if obj is _NotFound:
            if isinstance(index, EXPR.GetItemExpression):
                return index
            validated_index = self._validate_index(index)
            if validated_index is not index:
                index = validated_index
                # _processUnhashableIndex could have found a slice, or
                # _validate could have found an Ellipsis and returned a
                # slicer
                if index.__class__ is IndexedComponent_slice:
                    return index
                obj = self._data.get(index, _NotFound)
            #
            # Call the _getitem_when_not_present helper to retrieve/return
            # the default value
            #
            if obj is _NotFound:
                return self._getitem_when_not_present(index)

        return obj

    def __setitem__(self, index, val):
        #
        # Set the value: This relies on _setitem_when_not_present() to
        # insert the correct ComponentData into the _data dictionary
        # when it is not present and _setitem_impl to update an existing
        # entry.
        #
        # Note: it is important that we check _constructed is False and not
        # just evaluates to false: when constructing immutable Params,
        # _constructed will be None during the construction process when
        # setting the value is valid.
        #
        if self._constructed is False:
            self._not_constructed_error(index)

        try:
            obj = self._data.get(index, _NotFound)
        except TypeError:
            obj = _NotFound
            index = self._processUnhashableIndex(index)

        if obj is _NotFound:
            # If we didn't find the index in the data, then we need to
            # validate it against the underlying set (as long as
            # _processUnhashableIndex didn't return a slicer)
            if index.__class__ is not IndexedComponent_slice:
                index = self._validate_index(index)
        else:
            return self._setitem_impl(index, obj, val)
        #
        # Call the _setitem_impl helper to populate the _data
        # dictionary and set the value
        #
        # Note that we need to RECHECK the class against
        # IndexedComponent_slice, as _validate_index could have found
        # an Ellipsis (which is hashable) and returned a slicer
        #
        if index.__class__ is IndexedComponent_slice:
            # support "m.x[:,1] = 5" through a simple recursive call.
            #
            # Assert that this slice was just generated
            assert len(index._call_stack) == 1
            #
            # Note that the slicer will only slice over *existing*
            # entries, but they may not be in the data dictionary.  Make
            # a copy of the slicer items *before* we start iterating
            # over it in case the setter changes the _data dictionary.
            for idx, obj in list(index.expanded_items()):
                self._setitem_impl(idx, obj, val)
        else:
            obj = self._data.get(index, _NotFound)
            if obj is _NotFound:
                return self._setitem_when_not_present(index, val)
            else:
                return self._setitem_impl(index, obj, val)

    def __delitem__(self, index):
        if self._constructed is False:
            self._not_constructed_error(index)

        try:
            obj = self._data.get(index, _NotFound)
        except TypeError:
            obj = _NotFound
            index = self._processUnhashableIndex(index)

        if obj is _NotFound:
            if index.__class__ is not IndexedComponent_slice:
                index = self._validate_index(index)

        # this supports "del m.x[:,1]" through a simple recursive call
        if index.__class__ is IndexedComponent_slice:
            # Assert that this slice was just generated
            assert len(index._call_stack) == 1
            # Make a copy of the slicer items *before* we start
            # iterating over it (since we will be removing items!).
            for idx in list(index.expanded_keys()):
                del self[idx]
        else:
            # Handle the normal deletion operation
            if self.is_indexed():
                # Remove reference to this object
                self._data[index]._component = None
            del self._data[index]

    def _construct_from_rule_using_setitem(self):
        if self._rule is None:
            return
        index = None  # set so it is defined for scalars for `except:` below
        rule = self._rule
        block = self.parent_block()
        try:
            if rule.constant() and self.is_indexed():
                # A constant rule could return a dict-like thing or
                # matrix that we would then want to process with
                # Initializer().  If the rule actually returned a
                # constant, then this is just a little overhead.
                self._rule = rule = Initializer(
                    rule(block, None),
                    treat_sequences_as_mappings=False,
                    arg_not_specified=NOTSET,
                )

            if rule.contains_indices():
                # The index is coming in externally; we need to validate it
                for index in rule.indices():
                    self[index] = rule(block, index)
            elif not self.index_set().isfinite():
                # If the index is not finite, then we cannot iterate
                # over it.  Since the rule doesn't provide explicit
                # indices, then there is nothing we can do (the
                # assumption is that the user will trigger specific
                # indices to be created at a later time).
                pass
            elif rule.constant():
                # Slight optimization: if the initializer is known to be
                # constant, then only call the rule once.
                val = rule(block, None)
                for index in self.index_set():
                    self._setitem_when_not_present(index, val)
            else:
                for index in self.index_set():
                    self._setitem_when_not_present(index, rule(block, index))
        except:
            err = sys.exc_info()[1]
            logger.error(
                "Rule failed for %s '%s' with index %s:\n%s: %s"
                % (self.ctype.__name__, self.name, str(index), type(err).__name__, err)
            )
            raise

    def _not_constructed_error(self, idx):
        # Generate an error because the component is not constructed
        if not self.is_indexed():
            idx_str = ''
        elif idx.__class__ is tuple:
            idx_str = "[" + ",".join(str(i) for i in idx) + "]"
        else:
            idx_str = "[" + str(idx) + "]"
        raise ValueError(
            "Error retrieving component %s%s: The component has "
            "not been constructed." % (self.name, idx_str)
        )

    def _validate_index(self, idx):
        if not IndexedComponent._DEFAULT_INDEX_CHECKING_ENABLED:
            # Return whatever index was provided if the global flag dictates
            # that we should bypass all index checking and domain validation
            return idx

        # This is only called through __{get,set,del}item__, which has
        # already trapped unhashable objects.  Unfortunately, Python
        # 3.12 made slices hashable.  This means that slices will get
        # here and potentially be looked up in the index_set.  This will
        # cause problems with Any, where Any will happily return the
        # index as a valid set.  We will only validate the index for
        # non-Any sets.  Any will pass through so that normalize_index
        # can be called (which can generate the TypeError for slices)
        _any = isinstance(self._index_set, BASE.set._AnySet)
        if _any:
            validated_idx = _NotFound
        else:
            validated_idx = self._index_set.get(idx, _NotFound)
            if validated_idx is not _NotFound:
                # If the index is in the underlying index set, then return it
                #  Note: This check is potentially expensive (e.g., when the
                # indexing set is a complex set operation)!
                return validated_idx

        if normalize_index.flatten:
            # Now we normalize the index and check again.  Usually,
            # indices will be already be normalized, so we defer the
            # "automatic" call to normalize_index until now for the
            # sake of efficiency.
            normalized_idx = normalize_index(idx)
            if normalized_idx is not idx and not _any:
                if normalized_idx in self._data:
                    return normalized_idx
                if normalized_idx in self._index_set:
                    return normalized_idx
        else:
            normalized_idx = idx

        # There is the chance that the index contains an Ellipsis,
        # so we should generate a slicer
        if (
            normalized_idx.__class__ in slicer_types
            or normalized_idx.__class__ is tuple
            and any(_.__class__ in slicer_types for _ in normalized_idx)
        ):
            return self._processUnhashableIndex(normalized_idx)
        if _any:
            return idx
        #
        # Generate different errors, depending on the state of the index.
        #
        if not self.is_indexed():
            raise KeyError(
                "Cannot treat the scalar component '%s' "
                "as an indexed component" % (self.name,)
            )
        #
        # Raise an exception
        #
        raise KeyError(
            "Index '%s' is not valid for indexed component '%s'"
            % (normalized_idx, self.name)
        )

    def _processUnhashableIndex(self, idx):
        """Process a call to __getitem__ with unhashable elements

        There are three basic ways to get here:
          1) the index contains one or more slices or ellipsis
          2) the index contains an unhashable type (e.g., a Pyomo
             (Scalar)Component)
          3) the index contains an IndexTemplate
        """
        #
        # Iterate through the index and look for slices and constant
        # components
        #
        orig_idx = idx
        fixed = {}
        sliced = {}
        ellipsis = None
        #
        # Setup the slice template (in fixed)
        #
        if normalize_index.flatten:
            idx = normalize_index(idx)
        if idx.__class__ is not tuple:
            idx = (idx,)

        for i, val in enumerate(idx):
            if type(val) is slice:
                if (
                    val.start is not None
                    or val.stop is not None
                    or val.step is not None
                ):
                    raise IndexError(
                        "Indexed components can only be indexed with simple "
                        "slices: start and stop values are not allowed."
                    )
                else:
                    if ellipsis is None:
                        sliced[i] = val
                    else:
                        sliced[i - len(idx)] = val
                    continue

            if val is Ellipsis:
                if ellipsis is not None:
                    raise IndexError(
                        "Indexed components can only be indexed with simple "
                        "slices: the Pyomo wildcard slice (Ellipsis; "
                        "e.g., '...') can only appear once"
                    )
                ellipsis = i
                continue

            if hasattr(val, 'is_expression_type'):
                _num_val = val
                # Attempt to retrieve the numeric value .. if this
                # is a template expression generation, then it
                # should raise a TemplateExpressionError
                try:
                    val = EXPR.evaluate_expression(val, constant=True)
                except TemplateExpressionError:
                    #
                    # The index is a template expression, so return the
                    # templatized expression.
                    #
                    return EXPR.GetItemExpression((self,) + tuple(idx))

                except EXPR.NonConstantExpressionError:
                    #
                    # The expression contains an unfixed variable
                    #
                    raise RuntimeError(
                        """Error retrieving the value of an indexed item %s:
index %s is not a constant value.  This is likely not what you meant to
do, as if you later change the fixed value of the object this lookup
will not change.  If you understand the implications of using
non-constant values, you can get the current value of the object using
the value() function.""" % (self.name, i)
                    )

                except EXPR.FixedExpressionError:
                    #
                    # The expression contains a fixed variable
                    #
                    raise RuntimeError(
                        """Error retrieving the value of an indexed item %s:
index %s is a fixed but not constant value.  This is likely not what you
meant to do, as if you later change the fixed value of the object this
lookup will not change.  If you understand the implications of using
fixed but not constant values, you can get the current value using the
value() function.""" % (self.name, i)
                    )
                #
                # There are other ways we could get an exception such as
                # evaluating a Param / Var that is not initialized.
                # These exceptions will continue up the call stack.
                #

            # verify that the value is hashable
            hash(val)
            if ellipsis is None:
                fixed[i] = val
            else:
                fixed[i - len(idx)] = val

        if sliced or ellipsis is not None:
            slice_dim = len(idx)
            if ellipsis is not None:
                slice_dim -= 1
            if normalize_index.flatten:
                set_dim = self.dim()
            elif not self.is_indexed():
                # Scalar component.
                set_dim = 0
            else:
                set_dim = self.index_set().dimen
                if set_dim is None:
                    set_dim = 1

            structurally_valid = False
            if slice_dim == set_dim or set_dim is None:
                structurally_valid = True
            elif type(set_dim) is type:
                pass  # UnknownSetDimen
            elif ellipsis is not None and slice_dim < set_dim:
                structurally_valid = True
            elif set_dim == 0 and idx == (slice(None),):
                # If dim == 0 and idx is slice(None), the component was
                # a scalar passed a single slice. Since scalar components
                # can be accessed with a "1-dimensional" index of None,
                # this behavior is allowed.
                #
                # Note that x[...] is caught above, as slice_dim will be
                # 0 in that case
                structurally_valid = True

            if not structurally_valid:
                msg = (
                    "Index %s contains an invalid number of entries for "
                    "component '%s'. Expected %s, got %s."
                )
                if type(set_dim) is type:
                    set_dim = set_dim.__name__
                    msg += '\n    ' + '\n    '.join(
                        textwrap.wrap(
                            textwrap.dedent(
                                """
                                Slicing components relies on knowing the
                                underlying set dimensionality (even if the
                                dimensionality is None).  The underlying
                                component set ('%s') dimensionality has not been
                                determined (likely because it is an empty Set).
                                You can avoid this error by specifying the Set
                                dimensionality (with the 'dimen=' keyword)."""
                                % (self.index_set(),)
                            ).strip()
                        )
                    )
                raise IndexError(
                    msg
                    % (
                        IndexedComponent_slice._getitem_args_to_str(list(idx)),
                        self.name,
                        set_dim,
                        slice_dim,
                    )
                )
            return IndexedComponent_slice(self, fixed, sliced, ellipsis)
        elif len(idx) == len(fixed):
            if len(idx) == 1:
                return fixed[0]
            else:
                return tuple(fixed[i] for i in range(len(idx)))
        else:
            raise DeveloperError(
                "Unknown problem encountered when trying to retrieve "
                f"index '{orig_idx}' for component '{self.name}'"
            )

    def _getitem_when_not_present(self, index):
        """Returns/initializes a value when the index is not in the _data dict.

        Override this method if the component allows implicit member
        construction.  For classes that do not support a 'default' (at
        this point, everything except Param and Var), requesting
        _getitem_when_not_present will generate a KeyError (just like a
        normal dict).

        Implementations may assume that the index has already been validated
        and is a legitimate entry in the _data dict.

        """
        raise KeyError(index)

    def _setitem_impl(self, index, obj, value):
        """Perform the fundamental object value storage

        Components that want to implement a nonstandard storage mechanism
        should override this method.

        Implementations may assume that the index has already been
        validated and is a legitimate pre-existing entry in the _data
        dict.

        """
        if value is IndexedComponent.Skip:
            del self[index]
            return None
        else:
            obj.set_value(value)
        return obj

    def _setitem_when_not_present(self, index, value=_NotSpecified):
        """Perform the fundamental component item creation and storage.

        Components that want to implement a nonstandard storage mechanism
        should override this method.

        Implementations may assume that the index has already been
        validated and is a legitimate entry to add to the _data dict.
        """
        # If the value is "Skip" do not add anything
        if value is IndexedComponent.Skip:
            return None
        #
        # If we are a scalar, then idx will be None (_validate_index ensures
        # this)
        if index is None and not self.is_indexed():
            obj = self._data[index] = self
        else:
            obj = self._data[index] = self._ComponentDataClass(component=self)
        obj._index = index
        try:
            if value is not _NotSpecified:
                obj.set_value(value)
        except:
            self._data.pop(index, None)
            raise
        return obj

    def set_value(self, value):
        """Set the value of a scalar component."""
        if self.is_indexed():
            raise ValueError(
                "Cannot set the value for the indexed component '%s' "
                "without specifying an index value.\n"
                "\tFor example, model.%s[i] = value" % (self.name, self.name)
            )
        else:
            raise DeveloperError(
                "Derived component %s failed to define set_value() "
                "for scalar instances." % (self.__class__.__name__,)
            )

    def _pprint(self):
        """Print component information."""
        return (
            [
                ("Size", len(self)),
                ("Index", self._index_set if self.is_indexed() else None),
            ],
            self.items,
            ("Object",),
            lambda k, v: [type(v)],
        )

    def id_index_map(self):
        """
        Return an dictionary id->index for
        all ComponentData instances.
        """
        result = {}
        for index, component_data in self.items():
            result[id(component_data)] = index
        return result


class ActiveIndexedComponent(IndexedComponent, ActiveComponent):
    """
    This is the base class for all indexed modeling components
    whose data members are subclasses of ActiveComponentData, e.g.,
    can be activated or deactivated.

    The activate and deactivate methods activate both the
    component as well as all component data values.
    """

    def __init__(self, *args, **kwds):
        IndexedComponent.__init__(self, *args, **kwds)
        # Replicate the ActiveComponent.__init__() here.  We don't want
        # to use super, because that will run afoul of certain
        # assumptions for derived SimpleComponents' __init__()
        #
        # FIXME: eliminate multiple inheritance of SimpleComponents
        self._active = True

    def activate(self):
        """Set the active attribute to True"""
        super(ActiveIndexedComponent, self).activate()
        if self.is_indexed():
            for component_data in self.values():
                component_data.activate()

    def deactivate(self):
        """Set the active attribute to False"""
        super(ActiveIndexedComponent, self).deactivate()
        if self.is_indexed():
            for component_data in self.values():
                component_data.deactivate()


# Ideally, this would inherit from np.lib.mixins.NDArrayOperatorsMixin,
# but doing so overrides things like __contains__ in addition to the
# operators that we are interested in.
class IndexedComponent_NDArrayMixin:
    """Support using IndexedComponent with numpy.ndarray

    This IndexedComponent mixin class adds support for implicitly using
    the IndexedComponent as a term in an expression with numpy ndarray
    objects.

    """

    def __array__(self, dtype=None, copy=None):
        if dtype not in (None, object):
            raise ValueError(
                "Pyomo IndexedComponents can only be converted to NumPy "
                f"arrays with dtype=object (received {dtype=})"
            )
        if copy is not None and not copy:
            raise ValueError(
                "Pyomo IndexedComponents do not support conversion to NumPy "
                "arrays without generating a new array"
            )
        if not self.is_indexed():
            ans = _ndarray.NumericNDArray(shape=(1,), dtype=object)
            ans[0] = self
            return ans.reshape(())

        _dim = self.dim()
        if _dim is None:
            raise TypeError(
                "Cannot convert a non-dimensioned Pyomo IndexedComponent "
                "(%s) into a numpy array" % (self,)
            )
        bounds = self.index_set().bounds()
        if not isinstance(bounds[0], Sequence):
            bounds = ((bounds[0],), (bounds[1],))
        if any(b != 0 for b in bounds[0]):
            raise TypeError(
                "Cannot convert a Pyomo IndexedComponent "
                "(%s) with bounds [%s, %s] into a numpy array"
                % (self, bounds[0], bounds[1])
            )
        shape = tuple(b + 1 for b in bounds[1])
        ans = _ndarray.NumericNDArray(shape=shape, dtype=object)
        for k, v in self.items():
            ans[k] = v
        return ans

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return _ndarray.NumericNDArray.__array_ufunc__(
            None, ufunc, method, *inputs, **kwargs
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import pyomo.dae as dae


def make_gas_expansion_model(N=2):
    """
    This is the simplest model I could think of that has a
    subsystem with a non-trivial block triangularization.
    Something like a gas (somehow) undergoing a series
    of isentropic expansions.
    """
    m = pyo.ConcreteModel()
    m.streams = pyo.Set(initialize=range(N + 1))
    m.rho = pyo.Var(m.streams, initialize=1)
    m.P = pyo.Var(m.streams, initialize=1)
    m.F = pyo.Var(m.streams, initialize=1)
    m.T = pyo.Var(m.streams, initialize=1)

    m.R = pyo.Param(initialize=8.31)
    m.Q = pyo.Param(m.streams, initialize=1)
    m.gamma = pyo.Param(initialize=1.4 * m.R.value)

    def mbal(m, i):
        if i == 0:
            return pyo.Constraint.Skip
        else:
            return m.rho[i - 1] * m.F[i - 1] - m.rho[i] * m.F[i] == 0

    m.mbal = pyo.Constraint(m.streams, rule=mbal)

    def ebal(m, i):
        if i == 0:
            return pyo.Constraint.Skip
        else:
            return (
                m.rho[i - 1] * m.F[i - 1] * m.T[i - 1]
                + m.Q[i]
                - m.rho[i] * m.F[i] * m.T[i]
                == 0
            )

    m.ebal = pyo.Constraint(m.streams, rule=ebal)

    def expansion(m, i):
        if i == 0:
            return pyo.Constraint.Skip
        else:
            return m.P[i] / m.P[i - 1] - (m.rho[i] / m.rho[i - 1]) ** m.gamma == 0

    m.expansion = pyo.Constraint(m.streams, rule=expansion)

    def ideal_gas(m, i):
        return m.P[i] - m.rho[i] * m.R * m.T[i] == 0

    m.ideal_gas = pyo.Constraint(m.streams, rule=ideal_gas)

    return m


def make_dynamic_model(**disc_args):
    # Level control model
    m = pyo.ConcreteModel()
    m.time = dae.ContinuousSet(initialize=[0.0, 10.0])
    m.height = pyo.Var(m.time, initialize=1.0)
    m.flow_in = pyo.Var(m.time, initialize=1.0)
    m.flow_out = pyo.Var(m.time, initialize=0.5)
    m.dhdt = dae.DerivativeVar(m.height, wrt=m.time, initialize=0.0)

    m.area = pyo.Param(initialize=1.0)
    m.flow_const = pyo.Param(initialize=0.5)

    def diff_eqn_rule(m, t):
        return m.area * m.dhdt[t] - (m.flow_in[t] - m.flow_out[t]) == 0

    m.diff_eqn = pyo.Constraint(m.time, rule=diff_eqn_rule)

    def flow_out_rule(m, t):
        return m.flow_out[t] - (m.flow_const * pyo.sqrt(m.height[t])) == 0

    m.flow_out_eqn = pyo.Constraint(m.time, rule=flow_out_rule)

    default_disc_args = {"wrt": m.time, "nfe": 5, "scheme": "BACKWARD"}
    default_disc_args.update(disc_args)

    discretizer = pyo.TransformationFactory("dae.finite_difference")
    discretizer.apply_to(m, **default_disc_args)

    return m


def make_degenerate_solid_phase_model():
    """
    From the solid phase thermo package of a moving bed chemical looping
    combustion reactor. This example was first presented in [1]

    [1] Parker, R. Nonlinear programming strategies for dynamic models of
    chemical looping combustion reactors. Pres. AIChE Annual Meeting, 2020.

    """
    m = pyo.ConcreteModel()
    m.components = pyo.Set(initialize=[1, 2, 3])
    m.x = pyo.Var(m.components, initialize=1 / 3)
    m.flow_comp = pyo.Var(m.components, initialize=10)
    m.flow = pyo.Var(initialize=30)
    m.rho = pyo.Var(initialize=1)

    # These are rough approximations of the relevant equations, with the same
    # incidence.
    m.sum_eqn = pyo.Constraint(expr=sum(m.x[j] for j in m.components) - 1 == 0)
    m.holdup_eqn = pyo.Constraint(
        m.components, expr={j: m.x[j] * m.rho - 1 == 0 for j in m.components}
    )
    m.density_eqn = pyo.Constraint(
        expr=1 / m.rho - sum(1 / m.x[j] for j in m.components) == 0
    )
    m.flow_eqn = pyo.Constraint(
        m.components,
        expr={j: m.x[j] * m.flow - m.flow_comp[j] == 0 for j in m.components},
    )

    return m



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Author:  Gabe Hackebeil
# Purpose: For regression testing to ensure that the Pyomo
#          NL writer properly labels constraint ids in the "J"
#          section of the NL file when trivial constraints exist.
#          At the creation of this test, trivial constraints
#          (constraints with no variables) are being written to
#          the nl file as a feasibility check for the user.
#
#          This test model relies on the gjh_asl_json executable. It
#          will not solve if sent to a real optimizer.
#

from pyomo.environ import ConcreteModel, Var, Objective, Constraint, RangeSet

model = ConcreteModel()

n = 3

model.x = Var([(k, i) for k in range(1, n + 1) for i in range(k, n + 1)])


def obj_rule(model):
    return model.x[n, n]


model.obj = Objective(rule=obj_rule)


def var_bnd_rule(model, i):
    return (-1.0, model.x[1, i], 1.0)


model.var_bnd = Constraint(RangeSet(1, n), rule=var_bnd_rule)

model.x[1, 1] = 1.0
model.x[1, 1].fixed = True



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from .base_linear_solver_interface import IPLinearSolverInterface
from pyomo.contrib.pynumero.linalg.base import LinearSolverResults
from scipy.linalg import eigvals
from scipy.sparse import spmatrix
from pyomo.contrib.pynumero.sparse import BlockMatrix
import logging
import numpy as np
from pyomo.contrib.pynumero.linalg.scipy_interface import ScipyLU
from typing import Union


class ScipyInterface(ScipyLU, IPLinearSolverInterface):
    @classmethod
    def getLoggerName(cls):
        return 'ScipyLU'

    def __init__(self, compute_inertia=False):
        super(ScipyInterface, self).__init__()
        self._inertia = None
        self.compute_inertia = compute_inertia

        self.logger = logging.getLogger('scipy')
        self.logger.propagate = False

    def do_numeric_factorization(
        self, matrix: Union[spmatrix, BlockMatrix], raise_on_error: bool = True
    ) -> LinearSolverResults:
        res = super(ScipyInterface, self).do_numeric_factorization(
            matrix=matrix, raise_on_error=raise_on_error
        )

        if self.compute_inertia:
            eig = eigvals(matrix.toarray())
            pos_eig = np.count_nonzero((eig > 0))
            neg_eigh = np.count_nonzero((eig < 0))
            zero_eig = np.count_nonzero(eig == 0)
            self._inertia = (pos_eig, neg_eigh, zero_eig)

        return res

    def get_inertia(self):
        if self._inertia is None:
            raise RuntimeError(
                'The inertia was not computed during do_numeric_factorization. Set compute_inertia to True.'
            )
        return self._inertia



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.common.collections import ComponentMap
from pyomo.common.log import LoggingIntercept
from pyomo.core import Suffix, Var, Constraint, Piecewise, Block
from pyomo.core import Expression, Param
from pyomo.core.base.misc import apply_indexed_rule
from pyomo.core.base.block import IndexedBlock, SortComponents
from pyomo.dae import ContinuousSet, DAE_Error
from pyomo.common.formatting import tostr

from io import StringIO

logger = logging.getLogger('pyomo.dae')


def generate_finite_elements(ds, nfe):
    """
    This function first checks to see if the number of finite elements
    in the differential set is equal to nfe. If the number of finite
    elements is less than nfe, additional points will be generated. If
    the number of finite elements is greater than or equal to nfe the
    differential set will not be modified
    """
    if (len(ds) - 1) >= nfe:
        # In this case the differentialset already contains the
        # desired number or more than the desired number of finite
        # elements so no additional points are needed.
        return
    elif len(ds) == 2:
        # If only bounds have been specified on the differentialset we
        # generate the desired number of finite elements by
        # spreading them evenly over the interval
        step = (max(ds) - min(ds)) / float(nfe)
        tmp = min(ds) + step
        while round(tmp, 6) <= round((max(ds) - step), 6):
            ds.add(round(tmp, 6))
            tmp += step
        ds.set_changed(True)
        ds._fe = list(ds)
        return
    else:
        # This is the case where some points have been specified
        # inside of the bounds however the desired number of finite
        # elements has not been met. We first look at the step sizes
        # between the existing points. Then an additional point
        # is placed at the midpoint of the largest step. This
        # process is repeated until we have achieved the desired
        # number of finite elements. If there are multiple "largest steps"
        # the point will be placed at the first occurrence of the
        # largest step

        addpts = nfe - (len(ds) - 1)
        while addpts > 0:
            _add_point(ds)
            addpts -= 1
        ds.set_changed(True)
        ds._fe = list(ds)
        return


def _add_point(ds):
    sortds = list(ds)
    maxstep = sortds[1] - sortds[0]
    maxloc = 0
    for i in range(2, len(sortds)):
        if (sortds[i] - sortds[i - 1]) > maxstep:
            maxstep = sortds[i] - sortds[i - 1]
            maxloc = i - 1

    ds.add(round((sortds[maxloc] + maxstep / 2.0), 6))


def generate_colloc_points(ds, tau):
    """
    This function adds collocation points between the finite elements
    in the differential set
    """
    fes = list(ds)
    for i in range(1, len(fes)):
        h = fes[i] - fes[i - 1]
        for j in range(len(tau)):
            if tau[j] == 1 or tau[j] == 0:
                continue
            pt = fes[i - 1] + h * tau[j]
            pt = round(pt, 6)
            if pt not in ds:
                ds.add(pt)
                ds.set_changed(True)


def expand_components(block):
    """
    Loop over block components and try expanding them. If expansion fails
    then save the component and try again later. This function has some
    built-in robustness for block-hierarchical models with circular
    references but will not work for all cases.
    """

    # expansion_map is used to map components to the functions used to
    # expand them so that the update_contset_indexed_component function
    # logic only has to be called once even in the case where we have to
    # re-try expanding components due to circular references
    expansion_map = ComponentMap()
    redo_expansion = list()

    # Record the missing BlockData before expanding components. This is for
    # the case where a ContinuousSet indexed Block is used in a Constraint.
    # If the Constraint is expanded before the Block then the missing
    # BlockData will be added to the indexed Block but will not be
    # constructed correctly.
    for blk in block.component_objects(Block, descend_into=True):
        missing_idx = set(blk.index_set()) - set(blk._data.keys())
        if missing_idx:
            blk._dae_missing_idx = missing_idx

    # Wrap this whole process in a try block in order to ensure that errors
    # swallowed by the LoggingIntercept context below are re-raised if the
    # discretization encounters an error it isn't expecting.
    try:
        # Intercept logging to suppress Error messages arising from failed
        # constraint rules. These error messages get logged even though the
        # AttributeError causing the error is caught and handled by this
        # function when expanding discretized models. We maintain a stream
        # of the intercepted logging messages which will be printed if an
        # unexpected exception is raised.
        buf = StringIO()
        with LoggingIntercept(buf, 'pyomo.core', logging.ERROR):
            # Identify components that need to be expanded and try expanding
            # them
            for c in block.component_objects(
                descend_into=True, sort=SortComponents.declOrder
            ):
                try:
                    update_contset_indexed_component(c, expansion_map)
                except AttributeError:
                    redo_expansion.append(c)

            # Re-try expansion on any components that failed the first time.
            #  This is indicative of circular component references and not
            # expanding components in the correct order the first time
            # through.
            N = len(redo_expansion)
            while N:
                for i in range(N):
                    c = redo_expansion.pop()
                    try:
                        expansion_map[c](c)
                    except AttributeError:
                        redo_expansion.append(c)
                if len(redo_expansion) == N:
                    raise DAE_Error(
                        "Unable to fully discretize %s. Possible "
                        "circular references detected between "
                        "components %s. Reformulate your model to"
                        " remove circular references or apply a "
                        "discretization transformation before "
                        "linking blocks together." % (block, tostr(redo_expansion))
                    )

                N = len(redo_expansion)

    except Exception:
        logger.error(buf.getvalue())
        raise


def update_contset_indexed_component(comp, expansion_map):
    """
    Update any model components which are indexed by a ContinuousSet that
    has changed
    """

    # This implementation will *NOT* check for or update
    # components which use a ContinuousSet implicitly. ex) an
    # objective function which iterates through a ContinuousSet and
    # sums the squared error.  If you use a ContinuousSet implicitly
    # you must initialize it with every index you would like to have
    # access to!

    if comp.ctype is Suffix:
        return

    # Params indexed by a ContinuousSet should include an initialize
    # and/or default rule which will be called automatically when the
    # parameter value at a new point in the ContinuousSet is
    # requested. Therefore, no special processing is required for
    # Params.
    if comp.ctype is Param:
        return

    # Integral components are handled after every ContinuousSet has been
    # discretized. Import is deferred to here due to circular references.
    from pyomo.dae import Integral

    if comp.ctype is Integral:
        return

    # Skip components that do not have a 'dim' attribute. This assumes that
    # all components that could be indexed by a ContinuousSet have the 'dim'
    # attribute
    if not hasattr(comp, 'dim'):
        return

    # Components indexed by a ContinuousSet must have a dimension of at
    # least 1
    if comp.dim() == 0:
        return

    # Extract the indexing sets. Must treat components with a single
    # index separately from components with multiple indexing sets.
    temp = comp.index_set()
    indexset = list(comp.index_set().subsets())

    for s in indexset:
        if s.ctype == ContinuousSet and s.get_changed():
            if isinstance(comp, Var):  # Don't use the type() method here
                # because we want to catch DerivativeVar components as well
                # as Var components
                expansion_map[comp] = _update_var
                _update_var(comp)
            elif comp.ctype == Constraint:
                expansion_map[comp] = _update_constraint
                _update_constraint(comp)
            elif comp.ctype == Expression:
                expansion_map[comp] = _update_expression
                _update_expression(comp)
            elif isinstance(comp, Piecewise):
                expansion_map[comp] = _update_piecewise
                _update_piecewise(comp)
            elif comp.ctype == Block:
                expansion_map[comp] = _update_block
                _update_block(comp)
            else:
                raise TypeError(
                    "Found component %s of type %s indexed "
                    "by a ContinuousSet. Components of this type are "
                    "not currently supported by the automatic "
                    "discretization transformation in pyomo.dae. "
                    "Try adding the component to the model "
                    "after discretizing. Alert the pyomo developers "
                    "for more assistance." % (str(comp), comp.ctype)
                )


def _update_var(v):
    """
    This method will construct any additional indices in a variable
    resulting from the discretization of a ContinuousSet.
    """

    # Note: This is not required it is handled by the _default method on
    #       Var (which is now a IndexedComponent). However, it
    #       would be much slower to rely on that method to generate new
    #       VarData for a large number of new indices.
    new_indices = set(v.index_set()) - set(v._data.keys())
    for index in new_indices:
        v.add(index)


def _update_constraint(con):
    """
    This method will construct any additional indices in a constraint
    resulting from the discretization of a ContinuousSet.
    """
    con.to_dense_data()


def _update_expression(expre):
    """
    This method will construct any additional indices in an expression
    resulting from the discretization of a ContinuousSet.
    """
    expre.to_dense_data()


def _update_block(blk):
    """
    This method will construct any additional indices in a block
    resulting from the discretization of a ContinuousSet. For
    Block-derived components we check if the Block construct method has
    been overridden. If not then we update it like a regular block. If
    construct has been overridden then we try to call the component's
    update_after_discretization method. If the component hasn't
    implemented this method then we throw a warning and try to update it
    like a normal block. The issue, when construct is overridden, is that
    anything could be happening and we can't automatically assume that
    treating the block-derived component like a normal block will be
    sufficient to update it correctly.

    """

    # Check if Block construct method is overridden
    # getattr needed below for Python 2, 3 compatibility
    if blk.construct.__func__ is not getattr(
        IndexedBlock.construct, '__func__', IndexedBlock.construct
    ):
        # check for custom update function
        if hasattr(blk, 'update_after_discretization'):
            blk.update_after_discretization()
            return
        else:
            logger.warning(
                'DAE(misc): Attempting to apply a discretization '
                'transformation to the Block-derived component "%s". The '
                'component overrides the Block construct method but no '
                'update_after_discretization() function was found. Will '
                'attempt to update as a standard Block but user should verify '
                'that the component was expanded correctly. To suppress this '
                'warning, please provide an update_after_discretization() '
                'function on Block-derived components that override '
                'construct()' % blk.name
            )

    missing_idx = getattr(blk, '_dae_missing_idx', set([]))
    for idx in list(missing_idx):
        # Trigger block creation (including calling the Block's rule)
        blk[idx]

    # Remove book-keeping data after Block is discretized
    if hasattr(blk, '_dae_missing_idx'):
        del blk._dae_missing_idx


def _update_piecewise(pw):
    """
    This method will construct any additional indices in a Piecewise
    object resulting from the discretization of a ContinuousSet.
    """
    pw._constructed = False
    pw.construct()


def create_access_function(var):
    """
    This method returns a function that returns a component by calling
    it rather than indexing it
    """

    def _fun(*args):
        return var[args]

    return _fun


def create_partial_expression(scheme, expr, ind, loc):
    """
    This method returns a function which applies a discretization scheme
    to an expression along a particular indexing set. This is admittedly a
    convoluted looking implementation. The idea is that we only apply a
    discretization scheme to one indexing set at a time but we also want
    the function to be expanded over any other indexing sets.
    """

    def _fun(*args):
        return scheme(lambda i: expr(*(args[0:loc] + (i,) + args[loc + 1 :])), ind)

    return lambda *args: _fun(*args)(args[loc])


def add_discretization_equations(block, d):
    """
    Adds the discretization equations for DerivativeVar d to the Block block.
    Because certain indices will be valid for some discretization schemes and
    not others, we skip any constraints which raise an IndexError.
    """

    def _disc_eq(m, *args):
        try:
            return d[args] == d._expr(*args)
        except IndexError:
            return Constraint.Skip

    block.add_component(
        d.local_name + '_disc_eq', Constraint(d.index_set(), rule=_disc_eq)
    )


def add_continuity_equations(block, d, i, loc):
    """
    Adds continuity equations in the case that the polynomial basis function
    does not have a root at the finite element boundary
    """
    svar = d.get_state_var()
    nme = svar.local_name + '_' + i.local_name + '_cont_eq'
    if block.find_component(nme) is not None:
        return

    def _cont_exp(v, s):
        ncp = s.get_discretization_info()['ncp']
        afinal = s.get_discretization_info()['afinal']

        def _fun(i):
            tmp = list(s)
            idx = s.ord(i) - 1
            low = s.get_lower_element_boundary(i)
            if i != low or idx == 0:
                raise IndexError("list index out of range")
            low = s.get_lower_element_boundary(tmp[idx - 1])
            lowidx = s.ord(low) - 1
            return sum(v(tmp[lowidx + j]) * afinal[j] for j in range(ncp + 1))

        return _fun

    expr = create_partial_expression(_cont_exp, create_access_function(svar), i, loc)

    def _cont_eq(m, *args):
        try:
            return svar[args] == expr(*args)
        except IndexError:
            return Constraint.Skip

    block.add_component(nme, Constraint(d.index_set(), rule=_cont_eq))


def block_fully_discretized(b):
    """
    Checks to see if all ContinuousSets in a block have been discretized
    """

    for i in b.component_map(ContinuousSet).values():
        if 'scheme' not in i.get_discretization_info():
            return False
    return True


def get_index_information(var, ds):
    """
    This method will find the index location of the set ds in the var,
    return a list of the non_ds indices and return a function that can be
    used to access specific indices in var indexed by a ContinuousSet by
    specifying the finite element and collocation point. Users of this
    method should have already confirmed that ds is an indexing set of var
    and that it's a ContinuousSet
    """

    # Find index order of ContinuousSet in the variable
    indargs = []
    dsindex = 0

    if var.dim() != 1:
        indCount = 0
        for index in var.index_set().subsets():
            if isinstance(index, ContinuousSet):
                if index is ds:
                    dsindex = indCount
                else:
                    # If var is indexed by multiple ContinuousSets treat
                    # other ContinuousSets like a normal indexing set
                    indargs.append(index)
                indCount += 1  # A ContinuousSet must be one dimensional
            else:
                indargs.append(index)
                indCount += index.dimen

    if indargs == []:
        non_ds = (None,)
    elif len(indargs) > 1:
        non_ds = tuple(a for a in indargs)
    else:
        non_ds = (indargs[0],)

    if None in non_ds:
        tmpidx = (None,)
    elif len(non_ds) == 1:
        tmpidx = non_ds[0]
    else:
        tmpidx = non_ds[0].cross(*non_ds[1:])

    # Lambda function used to generate the desired index
    # more concisely
    idx = lambda n, i, k: _get_idx(dsindex, ds, n, i, k)

    info = dict()
    info['non_ds'] = tmpidx
    info['index function'] = idx
    return info


def _get_idx(l, ds, n, i, k):
    """
    This function returns the appropriate index for a variable
    indexed by a differential set. It's needed because the collocation
    constraints are indexed by finite element and collocation point
    however a ContinuousSet contains a list of all the discretization
    points and is not separated into finite elements and collocation
    points.
    """
    t = list(ds)
    tmp = ds.ord(ds._fe[i]) - 1
    tik = t[tmp + k]
    if n is None:
        return tik
    else:
        tmpn = n
        if not isinstance(n, tuple):
            tmpn = (n,)
    return tmpn[0:l] + (tik,) + tmpn[l:]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.common.collections import Bunch
from pyomo.opt import SolverFactory
from pyomo.opt.base.solvers import UnknownSolver

import pyomo.environ
from pyomo.solvers.plugins.solvers.GUROBI import GUROBISHELL
from pyomo.solvers.plugins.solvers.BARON import BARONSHELL
from pyomo.solvers.plugins.solvers.mosek_direct import MOSEKDirect

# ----------------------------------------------------------------

_test_solver_cases = {}


# ----------------------------------------------------------------

licensed_solvers_with_demo_mode = {'baron'}


#
# NOTE: we initialize the test case, since different
# interfaces may be used for the same "solver"
#
def initialize(**kwds):
    obj = Bunch(**kwds)
    #
    # Set obj.available
    #
    try:
        opt = SolverFactory(obj.name, solver_io=obj.io)
    except:
        opt = None

    if opt is None or isinstance(opt, UnknownSolver):
        obj.available = False
    elif not opt.available(exception_flag=False):
        obj.available = False
    elif hasattr(opt, 'executable') and opt.executable() is None:
        obj.available = False
    elif not opt.license_is_valid() and obj.name not in licensed_solvers_with_demo_mode:
        obj.available = False
    else:
        obj.available = True
    #
    # Set the limits for the solver's "demo" (unlicensed) mode:
    #   ( nVars, nCons, nNonZeros )
    obj.demo_limits = (None, None, None)
    if obj.available:
        if obj.name == "baron" and not opt.license_is_valid():
            obj.demo_limits = (10, 10, 50)
    #
    # Check capabilities, even if the solver is not available
    #
    if not (opt is None or isinstance(opt, UnknownSolver)):
        for _c in obj.capabilities:
            if not _c in opt._capabilities:
                raise ValueError(
                    "Solver %s does not support capability %s!" % (obj.name, _c)
                )
    #
    # Get version
    #
    if obj.available:
        obj.version = opt.version()
    return obj


def test_solver_cases(*args):
    """
    A function for accessing _test_solver_casess as global state
    """
    if len(_test_solver_cases) == 0:
        logging.disable(logging.WARNING)

        #
        # MOSEK
        #

        _mosek_capabilities = set(
            [
                'linear',
                'integer',
                'quadratic_objective',
                'quadratic_constraint',
                'conic_constraints',
            ]
        )

        _test_solver_cases['mosek', 'python'] = initialize(
            name='mosek',
            io='python',
            capabilities=_mosek_capabilities,
            import_suffixes=['dual', 'rc', 'slack'],
        )

        #
        # MOSEK Persistent
        #
        _test_solver_cases['mosek', 'persistent'] = initialize(
            name='mosek',
            io='persistent',
            capabilities=_mosek_capabilities,
            import_suffixes=['dual', 'rc', 'slack'],
        )

        #
        # CPLEX
        #

        _cplex_capabilities = set(
            [
                'linear',
                'integer',
                'quadratic_objective',
                'quadratic_constraint',
                'sos1',
                'sos2',
            ]
        )

        _test_solver_cases['cplex', 'lp'] = initialize(
            name='cplex',
            io='lp',
            capabilities=_cplex_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
        )

        _test_solver_cases['cplex', 'mps'] = initialize(
            name='cplex',
            io='mps',
            capabilities=_cplex_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
        )

        _test_solver_cases['cplex', 'nl'] = initialize(
            name='cplex',
            io='nl',
            capabilities=_cplex_capabilities,
            import_suffixes=['dual'],
        )

        _test_solver_cases['cplex', 'python'] = initialize(
            name='cplex',
            io='python',
            capabilities=_cplex_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
        )

        #
        # CPLEX PERSISTENT
        #

        _test_solver_cases['cplex_persistent', 'python'] = initialize(
            name='cplex_persistent',
            io='python',
            capabilities=_cplex_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
        )

        #
        # GAMS
        #

        _gams_capabilities = set(
            ['linear', 'integer', 'quadratic_objective', 'quadratic_constraint']
        )

        _test_solver_cases['gams', 'gms'] = initialize(
            name='gams',
            io='gms',
            capabilities=_gams_capabilities,
            import_suffixes=['dual', 'rc'],
        )

        _test_solver_cases['gams', 'python'] = initialize(
            name='gams',
            io='python',
            capabilities=_gams_capabilities,
            import_suffixes=['dual', 'rc'],
        )

        #
        # GAMS V2
        #

        _gams_v2_capabilities = set(['linear', 'integer'])

        _test_solver_cases['gams_v2', 'gms'] = initialize(
            name='gams_v2',
            io='gms',
            capabilities=_gams_v2_capabilities,
            import_suffixes=['dual', 'rc'],
        )

        _test_solver_cases['gams_v2', 'python'] = initialize(
            name='gams_v2',
            io='python',
            capabilities=_gams_v2_capabilities,
            import_suffixes=['dual', 'rc'],
        )

        #
        # GUROBI
        #
        # **NOTE: Gurobi does not handle quadratic constraints before
        #         Major Version 5
        #
        _gurobi_capabilities = set(
            [
                'linear',
                'integer',
                'quadratic_objective',
                'quadratic_constraint',
                'sos1',
                'sos2',
            ]
        )

        _test_solver_cases['gurobi', 'lp'] = initialize(
            name='gurobi',
            io='lp',
            capabilities=_gurobi_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
        )

        _test_solver_cases['gurobi', 'mps'] = initialize(
            name='gurobi',
            io='mps',
            capabilities=_gurobi_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
        )

        _test_solver_cases['gurobi', 'nl'] = initialize(
            name='gurobi',
            io='nl',
            capabilities=_gurobi_capabilities,
            options={'qcpdual': 1, 'simplex': 1},
            import_suffixes=['dual'],
        )

        _test_solver_cases['gurobi', 'python'] = initialize(
            name='gurobi',
            io='python',
            capabilities=_gurobi_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
        )

        #
        # Gurobi PERSISTENT
        #

        _test_solver_cases['gurobi_persistent', 'python'] = initialize(
            name='gurobi_persistent',
            io='python',
            capabilities=_gurobi_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
        )

        #
        # GLPK
        #
        _glpk_capabilities = set(['linear', 'integer'])

        if 'GLPKSHELL_old' in str(pyomo.solvers.plugins.solvers.GLPK.GLPK().__class__):
            glpk_import_suffixes = ['dual']
        else:
            glpk_import_suffixes = ['rc', 'dual']

        _test_solver_cases['glpk', 'lp'] = initialize(
            name='glpk',
            io='lp',
            capabilities=_glpk_capabilities,
            import_suffixes=glpk_import_suffixes,
        )

        _test_solver_cases['glpk', 'mps'] = initialize(
            name='glpk',
            io='mps',
            capabilities=_glpk_capabilities,
            import_suffixes=glpk_import_suffixes,
            io_options={"skip_objective_sense": True},
        )

        #
        # CBC
        #
        _cbc_lp_capabilities = set(['linear', 'integer'])

        _test_solver_cases['cbc', 'lp'] = initialize(
            name='cbc',
            io='lp',
            capabilities=_cbc_lp_capabilities,
            import_suffixes=['dual', 'rc'],
        )

        _cbc_nl_capabilities = set(['linear', 'integer', 'sos1', 'sos2'])

        _test_solver_cases['cbc', 'nl'] = initialize(
            name='cbc',
            io='nl',
            capabilities=_cbc_nl_capabilities,
            import_suffixes=['dual'],
        )

        # _cbc_mps_capabilities = set(['linear', 'integer', 'sos1', 'sos2'])

        # _test_solver_cases['cbc', 'mps'] = initialize(
        # name='cbc',
        # io='mps',
        # capabilities=_cbc_mps_capabilities,
        # import_suffixes=['dual', 'rc'])

        #
        # XPRESS
        #
        _xpress_capabilities = set(
            [
                'linear',
                'integer',
                'quadratic_objective',
                'quadratic_constraint',
                'sos1',
                'sos2',
            ]
        )

        _test_solver_cases['xpress', 'python'] = initialize(
            name='xpress_direct',
            io='python',
            capabilities=_xpress_capabilities,
            import_suffixes=['dual', 'rc', 'slack'],
            options={'bargapstop': 1e-9},
        )

        #
        # XPRESS PERSISTENT
        #

        _test_solver_cases['xpress_persistent', 'python'] = initialize(
            name='xpress_persistent',
            io='python',
            capabilities=_xpress_capabilities,
            import_suffixes=['slack', 'dual', 'rc'],
            options={'bargapstop': 1e-9},
        )

        #
        # IPOPT
        #
        _ipopt_capabilities = set(
            ['linear', 'quadratic_objective', 'quadratic_constraint']
        )

        _test_solver_cases['ipopt', 'nl'] = initialize(
            name='ipopt',
            io='nl',
            capabilities=_ipopt_capabilities,
            import_suffixes=['dual'],
        )

        #
        # SCIP
        #
        _scip_capabilities = set(
            [
                'linear',
                'integer',
                'quadratic_objective',
                'quadratic_constraint',
                'sos1',
                'sos2',
            ]
        )

        _test_solver_cases['scip', 'nl'] = initialize(
            name='scip', io='nl', capabilities=_scip_capabilities, import_suffixes=[]
        )

        #
        # CONOPT
        #
        _conopt_capabilities = set(
            [
                'linear',
                'integer',
                'quadratic_objective',
                'quadratic_constraint',
                'sos1',
                'sos2',
            ]
        )
        _test_solver_cases['conopt', 'nl'] = initialize(
            name='conopt',
            io='nl',
            capabilities=_conopt_capabilities,
            import_suffixes=[],
        )

        #
        # BARON
        #
        _baron_capabilities = set(
            ['linear', 'integer', 'quadratic_objective', 'quadratic_constraint']
        )

        _test_solver_cases['baron', 'bar'] = initialize(
            name='baron',
            io='bar',
            capabilities=_baron_capabilities,
            import_suffixes=['rc', 'dual'],
        )

        #
        # KNITROAMPL
        #
        _knitroampl_capabilities = set(
            ['linear', 'integer', 'quadratic_objective', 'quadratic_constraint']
        )

        _test_solver_cases['knitroampl', 'nl'] = initialize(
            name='knitroampl',
            io='nl',
            capabilities=_knitroampl_capabilities,
            import_suffixes=['dual'],
        )

        logging.disable(logging.NOTSET)

        #
        # CUOPT
        #
        _cuopt_capabilities = set(['linear', 'integer'])

        _test_solver_cases['cuopt', 'python'] = initialize(
            name='cuopt',
            io='python',
            capabilities=_cuopt_capabilities,
            import_suffixes=['rc', 'dual'],
        )

        #
        # Error Checks
        #
        for sc in _test_solver_cases.values():
            if sc.capabilities is None:
                sc.capabilities = set([])
            if sc.export_suffixes is None:
                sc.export_suffixes = []
            if sc.import_suffixes is None:
                sc.import_suffixes = []
            if sc.options is None:
                sc.options = {}
            if sc.io_options is None:
                sc.io_options = {}
            assert (sc.io is not None) and (type(sc.io) is str)
            assert type(sc.export_suffixes) in [list, tuple]
            assert type(sc.import_suffixes) in [list, tuple]
            assert type(sc.options) is dict
            for tag in sc.export_suffixes:
                assert type(tag) is str
            for tag in sc.import_suffixes:
                assert type(tag) is str
            assert type(sc.capabilities) is set
            for tag in sc.capabilities:
                assert type(tag) is str

    if len(args) == 0:
        return _test_solver_cases.keys()
    return _test_solver_cases[args]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests the integer to binary variable reformulation."""

import pyomo.common.unittest as unittest
from pyomo.environ import ConcreteModel, Var, Integers, value
from pyomo.environ import TransformationFactory as xfrm
from pyomo.common.log import LoggingIntercept

import logging
from io import StringIO


class TestIntToBinary(unittest.TestCase):
    """Tests integer to binary variable reformulation."""

    def test_int_to_binary(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers, bounds=(0, 5))
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.preprocessing', logging.INFO):
            xfrm('contrib.integer_to_binary').apply_to(m)
        self.assertIn(
            "Reformulating integer variables using the base2 strategy.",
            output.getvalue(),
        )
        reform_blk = m._int_to_binary_reform
        self.assertEqual(len(reform_blk.int_var_set), 1)
        reform_blk.new_binary_var[0, 0].value = 1
        reform_blk.new_binary_var[0, 1].value = 0
        reform_blk.new_binary_var[0, 2].value = 1
        m.x.value = 5
        self.assertEqual(value(reform_blk.integer_to_binary_constraint[0].body), 0)

    def test_int_to_binary_negative(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers, bounds=(-1, 1))
        xfrm('contrib.integer_to_binary').apply_to(m)
        reform_blk = m._int_to_binary_reform
        self.assertEqual(len(reform_blk.int_var_set), 1)
        reform_blk.new_binary_var[0, 0].value = 1
        reform_blk.new_binary_var[0, 1].value = 0
        m.x.value = 0
        # Check that 0 == -1 + 1 * 2^0 + 0 * 2^1
        self.assertEqual(value(reform_blk.integer_to_binary_constraint[0].body), 0)

    def test_integer_var_unbounded(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers)
        with self.assertRaises(ValueError):
            xfrm('contrib.integer_to_binary').apply_to(m)

    def test_no_integer(self):
        m = ConcreteModel()
        m.x = Var()
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.preprocessing', logging.INFO):
            xfrm('contrib.integer_to_binary').apply_to(m)

        expected_message = "Model has no free integer variables."
        self.assertIn(expected_message, output.getvalue())


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
import copy
import logging
import traceback
from pyomo.common.deprecation import deprecation_warning

logger = logging.getLogger(__name__)


pyomo_commands = []


def load_entry_points():
    import importlib.metadata

    ep_list = importlib.metadata.entry_points(group='pyomo.command')
    for ep in ep_list:
        try:
            pyomo_commands.append(ep.load())
        except:
            logger.error(
                f"Importing 'pyomo.command' entry point '{ep.name}' failed:\n"
                f"{traceback.format_exc(limit=1)}"
            )
            raise


def main(args=None):
    #
    # Load subcommands
    #
    from pyomo.scripting import pyomo_parser
    import pyomo.environ

    #
    # Load the pyomo.command entry points
    #
    if not pyomo_commands:
        load_entry_points()
    #
    # Parse the arguments
    #
    parser = pyomo_parser.get_parser()
    if args is None:
        args = copy.copy(sys.argv[1:])
    #
    # This is a hack to convert a command-line to a 'solve' subcommand
    #
    if not args:
        args.append('-h')
    if args[0][0] == '-':
        if args[0] not in ['-h', '--help', '--version']:
            deprecation_warning(
                "Running the 'pyomo' script with no subcommand is deprecated. "
                "Defaulting to 'pyomo solve'",
                version='6.5.0',
            )
            args = ['solve'] + args[0:]
    elif args[0] not in pyomo_parser.subparsers:
        deprecation_warning(
            "Running the 'pyomo' script with no subcommand is deprecated. "
            "Defaulting to 'pyomo solve'",
            version='6.5.0',
        )
        args = ['solve'] + args[0:]
    #
    # Process arguments
    #
    _options, _unparsed = parser.parse_known_args(args)
    #
    # Process the results
    #
    if _options.func.__code__.co_argcount == 1:
        #
        # If the execution function only accepts one argument, then we
        # create an exception if there are unparsed arguments.
        #
        if len(_unparsed) > 0:
            #
            # Re-parse the command-line to create an exception
            #
            parser.parse_args(_unparsed)
        retval = _options.func(_options)
    else:
        retval = _options.func(_options, _unparsed)
    return retval


def main_console_script():
    "This is the entry point for the main Pyomo script"
    # Note that we eat the retval data structure and only return the
    # process return code
    ans = main()
    try:
        return ans.errorcode
    except AttributeError:
        return ans


if __name__ == '__main__':
    sys.exit(main_console_script())



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import numpy.random as rnd
import pyomo.contrib.pynumero.examples.external_grey_box.param_est.models as pm
from pyomo.common.dependencies import pandas as pd


def generate_data(N, UA_mean, UA_std, seed=42):
    rnd.seed(seed)
    m = pyo.ConcreteModel()
    pm.build_single_point_model_pyomo_only(m)

    # dummy objective since this is a square problem
    m.obj = pyo.Objective(expr=1)

    # create the ipopt solver
    solver = pyo.SolverFactory('ipopt')

    data = {'run': [], 'Th_in': [], 'Tc_in': [], 'Th_out': [], 'Tc_out': []}
    for i in range(N):
        # draw a random value for the parameters
        ua = float(rnd.normal(UA_mean, UA_std))
        # draw a noisy value for the test input conditions
        Th_in = 100 + float(rnd.normal(0, 2))
        Tc_in = 30 + float(rnd.normal(0, 2))
        m.UA.fix(ua)
        m.Th_in.fix(Th_in)
        m.Tc_in.fix(Tc_in)

        status = solver.solve(m, tee=False)
        data['run'].append(i)
        data['Th_in'].append(pyo.value(m.Th_in))
        data['Tc_in'].append(pyo.value(m.Tc_in))
        data['Th_out'].append(pyo.value(m.Th_out))
        data['Tc_out'].append(pyo.value(m.Tc_out))

    return pd.DataFrame(data)


def generate_data_external(N, UA_mean, UA_std, seed=42):
    rnd.seed(seed)
    m = pyo.ConcreteModel()
    pm.build_single_point_model_external(m)

    # Add mutable parameters for the rhs of the equalities
    m.UA_spec = pyo.Param(initialize=200, mutable=True)
    m.Th_in_spec = pyo.Param(initialize=100, mutable=True)
    m.Tc_in_spec = pyo.Param(initialize=30, mutable=True)
    m.UA_spec_con = pyo.Constraint(expr=m.egb.inputs['UA'] == m.UA_spec)
    m.Th_in_spec_con = pyo.Constraint(expr=m.egb.inputs['Th_in'] == m.Th_in_spec)
    m.Tc_in_spec_con = pyo.Constraint(expr=m.egb.inputs['Tc_in'] == m.Tc_in_spec)

    # dummy objective since this is a square problem
    m.obj = pyo.Objective(expr=(m.egb.inputs['UA'] - m.UA_spec) ** 2)

    # create the ipopt solver
    solver = pyo.SolverFactory('cyipopt')

    data = {'run': [], 'Th_in': [], 'Tc_in': [], 'Th_out': [], 'Tc_out': []}
    for i in range(N):
        # draw a random value for the parameters
        UA = float(rnd.normal(UA_mean, UA_std))
        # draw a noisy value for the test input conditions
        Th_in = 100 + float(rnd.normal(0, 2))
        Tc_in = 30 + float(rnd.normal(0, 2))
        m.UA_spec.value = UA
        m.Th_in_spec.value = Th_in
        m.Tc_in_spec.value = Tc_in

        status = solver.solve(m, tee=False)
        data['run'].append(i)
        data['Th_in'].append(pyo.value(m.egb.inputs['Th_in']))
        data['Tc_in'].append(pyo.value(m.egb.inputs['Tc_in']))
        data['Th_out'].append(pyo.value(m.egb.inputs['Th_out']))
        data['Tc_out'].append(pyo.value(m.egb.inputs['Tc_out']))

    return pd.DataFrame(data)


if __name__ == '__main__':
    # df = generate_data(50, 200, 5)
    df = generate_data_external(50, 200, 5)
    df.to_csv('data.csv', index=False)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math

from pyomo.common.dependencies import numpy as numpy, numpy_available

if numpy_available:
    from numpy.testing import assert_array_almost_equal

import pyomo.environ as pyo
from pyomo.common import unittest

import pyomo.opt
from pyomo.contrib.alternative_solutions import (
    obbt_analysis_bounds_and_solutions,
    obbt_analysis,
)
import pyomo.contrib.alternative_solutions.tests.test_cases as tc

solvers = list(pyomo.opt.check_available_solvers("glpk", "gurobi", "appsi_gurobi"))
pytestmark = unittest.pytest.mark.parametrize("mip_solver", solvers)

timelimit = {"gurobi": "TimeLimit", "appsi_gurobi": "TimeLimit", "glpk": "tmlim"}


@unittest.pytest.mark.default
class TestOBBTUnit:

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_bad_solver(self, mip_solver):
        """
        Confirm that an exception is thrown with a bad solver name.
        """
        m = tc.get_2d_diamond_problem()
        try:
            obbt_analysis(m, solver="unknown_solver")
        except pyomo.common.errors.ApplicationError as e:
            pass

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_obbt_analysis(self, mip_solver):
        """
        Check that the correct bounds are found for a continuous problem.
        """
        m = tc.get_2d_diamond_problem()
        all_bounds = obbt_analysis(m, solver=mip_solver)
        assert all_bounds.keys() == m.continuous_bounds.keys()
        for var, bounds in all_bounds.items():
            assert_array_almost_equal(bounds, m.continuous_bounds[var])

    def test_obbt_error1(self, mip_solver):
        """
        ERROR: Cannot restrict variable list when warmstart is specified
        """
        m = tc.get_2d_diamond_problem()
        with unittest.pytest.raises(AssertionError):
            obbt_analysis_bounds_and_solutions(m, variables=[m.x], solver=mip_solver)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_obbt_some_vars(self, mip_solver):
        """
        Check that the correct bounds are found for a continuous problem.
        """
        m = tc.get_2d_diamond_problem()
        all_bounds, solns = obbt_analysis_bounds_and_solutions(
            m, variables=[m.x], warmstart=False, solver=mip_solver
        )
        assert len(all_bounds) == 1
        assert len(solns) == 2 * len(all_bounds) + 1
        for var, bounds in all_bounds.items():
            assert_array_almost_equal(bounds, m.continuous_bounds[var])

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_obbt_continuous(self, mip_solver):
        """
        Check that the correct bounds are found for a continuous problem.
        """
        m = tc.get_2d_diamond_problem()
        all_bounds, solns = obbt_analysis_bounds_and_solutions(m, solver=mip_solver)
        assert len(solns) == 2 * len(all_bounds) + 1
        assert all_bounds.keys() == m.continuous_bounds.keys()
        for var, bounds in all_bounds.items():
            assert_array_almost_equal(bounds, m.continuous_bounds[var])

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_mip_rel_objective(self, mip_solver):
        """
        Check that relative mip gap constraints are added for a mip with indexed vars and constraints
        """
        m = tc.get_indexed_pentagonal_pyramid_mip()
        all_bounds, solns = obbt_analysis_bounds_and_solutions(
            m, rel_opt_gap=0.5, solver=mip_solver
        )
        assert len(solns) == 2 * len(all_bounds) + 1
        assert m._obbt.optimality_tol_rel.lb == unittest.pytest.approx(2.5)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_mip_abs_objective(self, mip_solver):
        """
        Check that absolute mip gap constraints are added
        """
        m = tc.get_pentagonal_pyramid_mip()
        all_bounds, solns = obbt_analysis_bounds_and_solutions(
            m, abs_opt_gap=1.99, solver=mip_solver
        )
        assert len(solns) == 2 * len(all_bounds) + 1
        assert m._obbt.optimality_tol_abs.lb == unittest.pytest.approx(3.01)

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_obbt_warmstart(self, mip_solver):
        """
        Check that warmstarting works.
        """
        m = tc.get_2d_diamond_problem()
        m.x.value = 0
        m.y.value = 0
        all_bounds, solns = obbt_analysis_bounds_and_solutions(
            m, solver=mip_solver, warmstart=True, tee=False
        )
        assert len(solns) == 2 * len(all_bounds) + 1
        assert all_bounds.keys() == m.continuous_bounds.keys()
        for var, bounds in all_bounds.items():
            assert_array_almost_equal(bounds, m.continuous_bounds[var])

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_obbt_mip(self, mip_solver):
        """
        Check that bound tightening only occurs for continuous variables
        that can be tightened.
        """
        m = tc.get_bloated_pentagonal_pyramid_mip()
        all_bounds, solns = obbt_analysis_bounds_and_solutions(
            m, solver=mip_solver, tee=False
        )
        assert len(solns) == 2 * len(all_bounds) + 1
        bounds_tightened = False
        bounds_not_tightned = False
        for var, bounds in all_bounds.items():
            if bounds[0] > var.lb:
                bounds_tightened = True
            else:
                bounds_not_tightened = True
            if bounds[1] < var.ub:
                bounds_tightened = True
            else:
                bounds_not_tightened = True
        assert bounds_tightened
        assert bounds_not_tightened

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_obbt_unbounded(self, mip_solver):
        """
        Check that the correct bounds are found for an unbounded problem.
        """
        m = tc.get_2d_unbounded_problem()
        all_bounds, solns = obbt_analysis_bounds_and_solutions(m, solver=mip_solver)
        assert all_bounds.keys() == m.continuous_bounds.keys()
        num = 1
        for var, bounds in all_bounds.items():
            if not math.isinf(bounds[0]):
                num += 1
            if not math.isinf(bounds[1]):
                num += 1
            assert_array_almost_equal(bounds, m.continuous_bounds[var])
        assert len(solns) == num

    @unittest.skipIf(not numpy_available, "Numpy not installed")
    def test_bound_tightening(self, mip_solver):
        """
        Check that the correct bounds are found for a discrete problem where
        more restrictive bounds are implied by the constraints.
        """
        m = tc.get_implied_bound_ip()
        all_bounds, solns = obbt_analysis_bounds_and_solutions(m, solver=mip_solver)
        assert len(solns) == 2 * len(all_bounds) + 1
        assert all_bounds.keys() == m.var_bounds.keys()
        for var, bounds in all_bounds.items():
            assert_array_almost_equal(bounds, m.var_bounds[var])

    @unittest.skipIf(True, "Ignoring fragile test for solver timeout.")
    def test_no_time(self, mip_solver):
        """
        Check that the correct bounds are found for a discrete problem where
        more restrictive bounds are implied by the constraints.
        """
        m = tc.get_implied_bound_ip()
        with unittest.pytest.raises(RuntimeError):
            obbt_analysis_bounds_and_solutions(
                m, solver=mip_solver, solver_options={timelimit[mip_solver]: 0}
            )

    def test_bound_refinement(self, mip_solver):
        """
        Check that the correct bounds are found for a discrete problem where
        more restrictive bounds are implied by the constraints and constraints
        are added.
        """
        m = tc.get_implied_bound_ip()
        all_bounds, solns = obbt_analysis_bounds_and_solutions(
            m, solver=mip_solver, refine_discrete_bounds=True
        )
        assert len(solns) == 2 * len(all_bounds) + 1
        for var, bounds in all_bounds.items():
            if m.var_bounds[var][0] > var.lb:
                match = False
                for idx in m._obbt.bound_constraints:
                    const = m._obbt.bound_constraints[idx]
                    if var is const.body and bounds[0] == const.lb:
                        match = True
                        break
                assert match, "Constraint not found for {} lower bound {}".format(
                    var, bounds[0]
                )
            if m.var_bounds[var][1] < var.ub:
                match = False
                for idx in m._obbt.bound_constraints:
                    const = m._obbt.bound_constraints[idx]
                    if var is const.body and bounds[1] == const.ub:
                        match = True
                        break
                assert match, "Constraint not found for {} upper bound {}".format(
                    var, bounds[1]
                )

    def test_obbt_infeasible(self, mip_solver):
        """
        Check that code catches cases where the problem is infeasible.
        """
        m = tc.get_2d_diamond_problem()
        m.infeasible_constraint = pyo.Constraint(expr=m.x >= 10)
        with unittest.pytest.raises(Exception):
            obbt_analysis_bounds_and_solutions(m, solver=mip_solver)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

import pyomo.environ
from pyomo.core import ConcreteModel, Var, Objective, Constraint, NonNegativeReals
from pyomo.opt import SolverFactory

try:
    import cplex

    cplexpy_available = True
except ImportError:
    cplexpy_available = False


@unittest.skipIf(not cplexpy_available, "The 'cplex' python bindings are not available")
@unittest.pytest.mark.solver("cplex_persistent")
class TestQuadraticObjective(unittest.TestCase):
    def test_quadratic_objective_is_set(self):
        model = ConcreteModel()
        model.X = Var(bounds=(-2, 2))
        model.Y = Var(bounds=(-2, 2))
        model.O = Objective(expr=model.X**2 + model.Y**2)
        model.C1 = Constraint(expr=model.Y >= 2 * model.X - 1)
        model.C2 = Constraint(expr=model.Y >= -model.X + 2)
        opt = SolverFactory("cplex_persistent")
        opt.set_instance(model)
        opt.solve()

        self.assertAlmostEqual(model.X.value, 1, places=3)
        self.assertAlmostEqual(model.Y.value, 1, places=3)

        del model.O
        model.O = Objective(expr=model.X**2)
        opt.set_objective(model.O)
        opt.solve()
        self.assertAlmostEqual(model.X.value, 0, places=3)
        self.assertAlmostEqual(model.Y.value, 2, places=3)

    def test_add_column(self):
        m = ConcreteModel()
        m.x = Var(within=NonNegativeReals)
        m.c = Constraint(expr=(0, m.x, 1))
        m.obj = Objective(expr=-m.x)

        opt = SolverFactory('cplex_persistent')
        opt.set_instance(m)
        opt.solve()
        self.assertAlmostEqual(m.x.value, 1)

        m.y = Var(within=NonNegativeReals)

        opt.add_column(m, m.y, -3, [m.c], [2])
        opt.solve()

        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 0.5)

    def test_add_column_exceptions(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=(0, m.x, 1))
        m.ci = Constraint([1, 2], rule=lambda m, i: (0, m.x, i + 1))
        m.cd = Constraint(expr=(0, -m.x, 1))
        m.cd.deactivate()
        m.obj = Objective(expr=-m.x)

        opt = SolverFactory('cplex_persistent')

        # set_instance not called
        self.assertRaises(RuntimeError, opt.add_column, m, m.x, 0, [m.c], [1])

        opt.set_instance(m)

        m2 = ConcreteModel()
        m2.y = Var()
        m2.c = Constraint(expr=(0, m.x, 1))

        # different model than attached to opt
        self.assertRaises(RuntimeError, opt.add_column, m2, m2.y, 0, [], [])
        # pyomo var attached to different model
        self.assertRaises(RuntimeError, opt.add_column, m, m2.y, 0, [], [])

        z = Var()
        # pyomo var floating
        self.assertRaises(RuntimeError, opt.add_column, m, z, -2, [m.c, z], [1])

        m.y = Var()
        # len(coefficients) == len(constraints)
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c], [1, 2])
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c, z], [1])

        # add indexed constraint
        self.assertRaises(AttributeError, opt.add_column, m, m.y, -2, [m.ci], [1])
        # add something not a ConstraintData
        self.assertRaises(AttributeError, opt.add_column, m, m.y, -2, [m.x], [1])

        # constraint not on solver model
        self.assertRaises(KeyError, opt.add_column, m, m.y, -2, [m2.c], [1])

        # inactive constraint
        self.assertRaises(KeyError, opt.add_column, m, m.y, -2, [m.cd], [1])

        opt.add_var(m.y)
        # var already in solver model
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c], [1])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import json
import os
from os.path import abspath, dirname, join

import pyomo.common.unittest as unittest

from pyomo.common.fileutils import this_file_dir
from pyomo.common.log import LoggingIntercept
from pyomo.common.tempfiles import TempfileManager

import pyomo.opt
from pyomo.core import (
    ConcreteModel,
    RangeSet,
    Var,
    Param,
    Objective,
    ConstraintList,
    value,
    minimize,
)

currdir = this_file_dir()
deleteFiles = True

ipopt_available = False


class Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        global ipopt_available
        import pyomo.environ
        from pyomo.solvers.tests.solvers import test_solver_cases as _test_solver_cases

        ipopt_available = _test_solver_cases('ipopt', 'nl').available

    def setUp(self):
        if not ipopt_available:
            self.skipTest("The 'ipopt' command is not available")
        TempfileManager.push()

        self.asl = pyomo.opt.SolverFactory('asl:ipopt', keepfiles=True)
        self.ipopt = pyomo.opt.SolverFactory('ipopt', keepfiles=True)

        # The sisser CUTEr instance
        # Formulated in Pyomo by Carl D. Laird, Daniel P. Word, Brandon
        #     C. Barrera and Saumyajyoti Chaudhuri
        # Taken from:

        #   Source:
        #   F.S. Sisser,
        #   "Elimination of bounds in optimization problems by transforming
        #   variables",
        #   Mathematical Programming 20:110-121, 1981.

        #   See also Buckley#216 (p. 91)

        #   SIF input: Ph. Toint, Dec 1989.

        #   classification OUR2-AN-2-0

        sisser_instance = ConcreteModel()

        sisser_instance.N = RangeSet(1, 2)
        sisser_instance.xinit = Param(sisser_instance.N, initialize={1: 1.0, 2: 0.1})

        def fa(model, i):
            return value(model.xinit[i])

        sisser_instance.x = Var(sisser_instance.N, initialize=fa)

        def f(model):
            return (
                3 * model.x[1] ** 4
                - 2 * (model.x[1] * model.x[2]) ** 2
                + 3 * model.x[2] ** 4
            )

        sisser_instance.f = Objective(rule=f, sense=minimize)

        self.sisser_instance = sisser_instance

    def tearDown(self):
        TempfileManager.pop(remove=deleteFiles or self.currentTestPassed())

    def compare_json(self, file1, file2):
        with open(file1, 'r') as out, open(file2, 'r') as txt:
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-7, allow_second_superset=True
            )

    def test_version_asl(self):
        self.assertTrue(self.asl.version() is not None)
        self.assertTrue(type(self.asl.version()) is tuple)
        self.assertEqual(len(self.asl.version()), 4)

    def test_version_ipopt(self):
        self.assertTrue(self.ipopt.version() is not None)
        self.assertTrue(type(self.ipopt.version()) is tuple)
        self.assertEqual(len(self.ipopt.version()), 4)

    def test_asl_solve_from_nl(self):
        # Test ipopt solve from nl file
        _log = TempfileManager.create_tempfile(".test_ipopt.log")
        results = self.asl.solve(
            join(currdir, "sisser.pyomo.nl"), logfile=_log, suffixes=['.*']
        )
        # We don't want the test to care about which Ipopt version we are using
        results.Solution(0).Message = "Ipopt"
        results.Solver.Message = "Ipopt"
        _out = TempfileManager.create_tempfile(".test_ipopt.txt")
        results.write(filename=_out, times=False, format='json')
        self.compare_json(_out, join(currdir, "test_solve_from_nl.baseline"))

    def test_ipopt_solve_from_nl(self):
        # Test ipopt solve from nl file
        _log = TempfileManager.create_tempfile(".test_ipopt.log")
        results = self.ipopt.solve(
            join(currdir, "sisser.pyomo.nl"), logfile=_log, suffixes=['.*']
        )
        # We don't want the test to care about which Ipopt version we are using
        results.Solution(0).Message = "Ipopt"
        results.Solver.Message = "Ipopt"
        _out = TempfileManager.create_tempfile(".test_ipopt.txt")
        results.write(filename=_out, times=False, format='json')
        self.compare_json(_out, join(currdir, "test_solve_from_nl.baseline"))

    def test_asl_solve_from_instance(self):
        # Test ipopt solve from a pyomo instance and load the solution
        results = self.asl.solve(self.sisser_instance, suffixes=['.*'])
        # We don't want the test to care about which Ipopt version we are using
        self.sisser_instance.solutions.store_to(results)
        results.Solution(0).Message = "Ipopt"
        results.Solver.Message = "Ipopt"
        _out = TempfileManager.create_tempfile(".test_ipopt.txt")
        results.write(filename=_out, times=False, format='json')
        self.compare_json(_out, join(currdir, "test_solve_from_instance.baseline"))
        # self.sisser_instance.load_solutions(results)

    def test_ipopt_solve_from_instance(self):
        # Test ipopt solve from a pyomo instance and load the solution
        results = self.ipopt.solve(self.sisser_instance, suffixes=['.*'])
        # We don't want the test to care about which Ipopt version we are using
        self.sisser_instance.solutions.store_to(results)
        results.Solution(0).Message = "Ipopt"
        results.Solver.Message = "Ipopt"
        _out = TempfileManager.create_tempfile(".test_ipopt.txt")
        results.write(filename=_out, times=False, format='json')
        self.compare_json(_out, join(currdir, "test_solve_from_instance.baseline"))
        # self.sisser_instance.load_solutions(results)

    def test_ipopt_solve_from_instance_OF_options(self):
        with self.assertRaises(ValueError):
            # using OF_ options AND option_file_name
            # is not allowed
            self.ipopt.solve(
                self.sisser_instance,
                suffixes=['.*'],
                options={"OF_mu_init": 0.1, "option_file_name": "junk.opt"},
            )
        # Creating a dummy ipopt.opt file in the cwd
        # will cover the code that prints a warning
        _cwd = os.getcwd()
        tmpdir = TempfileManager.create_tempdir()
        try:
            os.chdir(tmpdir)
            # create an empty ipopt.opt file
            open(join(tmpdir, 'ipopt.opt'), "w").close()
            # Test ipopt solve from a pyomo instance and load the solution
            with LoggingIntercept() as LOG:
                results = self.ipopt.solve(
                    self.sisser_instance, suffixes=['.*'], options={"OF_mu_init": 0.1}
                )
            self.assertRegex(
                LOG.getvalue().replace("\n", " "),
                r"A file named (.*) exists in the current working "
                r"directory, but Ipopt options file options \(i.e., "
                r"options that start with 'OF_'\) were provided. The "
                r"options file \1 will be ignored.",
            )
        finally:
            os.chdir(_cwd)

        # We don't want the test to care about which Ipopt version we are using
        self.sisser_instance.solutions.store_to(results)
        results.Solution(0).Message = "Ipopt"
        results.Solver.Message = "Ipopt"
        _out = TempfileManager.create_tempfile(".test_ipopt.txt")
        results.write(filename=_out, times=False, format='json')
        self.compare_json(_out, join(currdir, "test_solve_from_instance.baseline"))
        # self.sisser_instance.load_solutions(results)

    def test_bad_dof(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.c = ConstraintList()
        m.c.add(m.x + m.y == 1)
        m.c.add(m.x - m.y == 0)
        m.c.add(2 * m.x - 3 * m.y == 1)
        res = self.ipopt.solve(m)
        self.assertEqual(str(res.solver.status), "warning")
        self.assertEqual(str(res.solver.termination_condition), "other")
        self.assertTrue("Too few degrees of freedom" in res.solver.message)


if __name__ == "__main__":
    deleteFiles = False
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from typing import Dict, Any, List, Sequence, Optional, Mapping, NoReturn

from pyomo.core.base import Var
from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.var import VarData
from pyomo.common.collections import ComponentMap
from pyomo.core.staleflag import StaleFlagManager
from pyomo.repn.plugins.gams_writer_v2 import GAMSWriterInfo
from pyomo.contrib.solver.common.solution_loader import SolutionLoaderBase
from pyomo.contrib.solver.common.util import (
    NoDualsError,
    NoSolutionError,
    NoReducedCostsError,
)


class GDXFileData:
    """
    Defines the data types found within a .gdx file
    """

    def __init__(self) -> None:
        self.primals: List[float] = []
        self.duals: List[float] = []
        self.var_suffixes: Dict[str, Dict[int, Any]] = {}
        self.con_suffixes: Dict[str, Dict[Any]] = {}
        self.obj_suffixes: Dict[str, Dict[int, Any]] = {}
        self.problem_suffixes: Dict[str, List[Any]] = {}
        self.other: List[str] = []


class GMSSolutionLoader(SolutionLoaderBase):
    """
    Loader for solvers that create .gms files (e.g., gams)
    """

    def __init__(self, gdx_data: GDXFileData, gms_info: GAMSWriterInfo) -> None:
        self._gdx_data = gdx_data
        self._gms_info = gms_info

    def load_vars(self, vars_to_load: Optional[Sequence[VarData]] = None) -> NoReturn:
        if self._gms_info is None:
            raise NoSolutionError()
        if self._gdx_data is None:
            assert len(self._gms_info.var_symbol_map.bySymbol) == 0
        else:
            for sym, obj in self._gms_info.var_symbol_map.bySymbol.items():
                level = self._gdx_data[sym][0]
                obj.set_value(level, skip_validation=True)

        StaleFlagManager.mark_all_as_stale(delayed=True)

    def get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        if self._gms_info is None:
            raise NoSolutionError()
        val_map = {}
        if self._gdx_data is None:
            assert len(self._gms_info.var_symbol_map.bySymbol) == 0
        else:
            for sym, obj in self._gms_info.var_symbol_map.bySymbol.items():
                val_map[id(obj)] = self._gdx_data[sym][0]

        res = ComponentMap()
        if vars_to_load is None:
            vars_to_load = self._gms_info.var_symbol_map.bySymbol.values()

        for obj in vars_to_load:
            res[obj] = val_map[id(obj)]

        return res

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        if self._gms_info is None:
            raise NoDualsError()
        if self._gdx_data is None:
            raise NoDualsError()

        con_map = {}
        for sym, obj in self._gms_info.con_symbol_map.bySymbol.items():
            con_map[id(obj)] = self._gdx_data[sym][1]

        for sym, obj in self._gms_info.con_symbol_map.aliases.items():
            if self._gdx_data[sym][1] != 0:
                con_map[id(obj)] = self._gdx_data[sym][1]

        res = ComponentMap()
        if cons_to_load is None:
            cons_to_load = self._gms_info.con_symbol_map.bySymbol.values()

        for obj in cons_to_load:
            res[obj] = con_map[id(obj)]

        return res

    def get_reduced_costs(self, vars_to_load=None):
        if self._gms_info is None:
            raise NoReducedCostsError()
        if self._gdx_data is None:
            raise NoReducedCostsError()

        var_map = {}
        for sym, obj in self._gms_info.var_symbol_map.bySymbol.items():
            var_map[id(obj)] = self._gdx_data[sym][1]

        res = ComponentMap()
        if vars_to_load is None:
            vars_to_load = self._gms_info.var_symbol_map.bySymbol.values()

        for obj in vars_to_load:
            res[obj] = var_map[id(obj)]

        return res



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests the variable aggregation module."""

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentSet
from pyomo.contrib.preprocessing.plugins.var_aggregator import (
    _build_equality_set,
    _get_equality_linked_variables,
    max_if_not_None,
    min_if_not_None,
)
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    ConstraintList,
    maximize,
    Objective,
    RangeSet,
    Reals,
    SolverFactory,
    TransformationFactory,
    Var,
)


class TestVarAggregate(unittest.TestCase):
    """Tests variable aggregation."""

    def build_model(self):
        m = ConcreteModel()
        m.v1 = Var(initialize=1, bounds=(1, 8))
        m.v2 = Var(initialize=2, bounds=(0, 3))
        m.v3 = Var(initialize=3, bounds=(-7, 4))
        m.v4 = Var(initialize=4, bounds=(2, 6))
        m.c1 = Constraint(expr=m.v1 == m.v2)
        m.c2 = Constraint(expr=m.v2 == m.v3)
        m.c3 = Constraint(expr=m.v3 == m.v4)
        m.v2.fix()

        m.s = RangeSet(5)
        m.x = Var(m.s, initialize=5)
        m.c = Constraint(m.s)
        m.c.add(1, expr=m.x[1] == m.x[3])
        m.c.add(2, expr=m.x[2] == m.x[4])
        m.c.add(3, expr=m.x[2] == m.x[3])
        m.c.add(4, expr=m.x[1] == 1)
        m.c.add(5, expr=(2, m.x[5], 3))

        m.y = Var([1, 2], initialize={1: 3, 2: 4})
        m.c_too = Constraint(expr=m.y[1] == m.y[2])

        m.z1 = Var()
        m.z2 = Var()
        m.ignore_me = Constraint(expr=m.y[1] + m.z1 + m.z2 <= 0)
        m.ignore_me_too = Constraint(expr=m.y[1] * m.y[2] == 0)
        m.multiple = Constraint(expr=m.y[1] == 2 * m.y[2])

        return m

    def test_aggregate_fixed_var_diff_values(self):
        m = ConcreteModel()
        m.s = RangeSet(3)
        m.v = Var(m.s, bounds=(0, 5))
        m.c = ConstraintList()
        m.c.add(expr=m.v[1] == m.v[2])
        m.c.add(expr=m.v[2] == m.v[3])
        m.c.add(expr=m.v[1] == 1)
        m.c.add(expr=m.v[3] == 3)
        with self.assertRaises(ValueError):
            TransformationFactory('contrib.aggregate_vars').apply_to(m)

    def test_fixed_var_out_of_bounds_lb(self):
        m = ConcreteModel()
        m.s = RangeSet(2)
        m.v = Var(m.s, bounds=(0, 5))
        m.c = ConstraintList()
        m.c.add(expr=m.v[1] == m.v[2])
        m.c.add(expr=m.v[1] == -1)
        with self.assertRaises(ValueError):
            TransformationFactory('contrib.aggregate_vars').apply_to(m)

    def test_fixed_var_out_of_bounds_ub(self):
        m = ConcreteModel()
        m.s = RangeSet(2)
        m.v = Var(m.s, bounds=(0, 5))
        m.c = ConstraintList()
        m.c.add(expr=m.v[1] == m.v[2])
        m.c.add(expr=m.v[1] == 6)
        with self.assertRaises(ValueError):
            TransformationFactory('contrib.aggregate_vars').apply_to(m)

    def test_do_not_transform_deactivated_constraints(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.c1 = Constraint(expr=m.x == m.y)
        m.c2 = Constraint(expr=(2, m.x, 3))
        m.c3 = Constraint(expr=m.x == 0)
        m.c3.deactivate()
        TransformationFactory('contrib.aggregate_vars').apply_to(m)
        self.assertIs(m.c2.body, m._var_aggregator_info.z[1])
        self.assertIs(m.c3.body, m.x)

    def test_equality_linked_variables(self):
        """Test for equality-linked variable detection."""
        m = self.build_model()
        self.assertEqual(_get_equality_linked_variables(m.c1), ())
        self.assertEqual(_get_equality_linked_variables(m.c2), ())
        c3 = _get_equality_linked_variables(m.c3)
        self.assertIn(m.v3, ComponentSet(c3))
        self.assertIn(m.v4, ComponentSet(c3))
        self.assertEqual(len(c3), 2)
        self.assertEqual(_get_equality_linked_variables(m.ignore_me), ())
        self.assertEqual(_get_equality_linked_variables(m.ignore_me_too), ())
        self.assertEqual(_get_equality_linked_variables(m.multiple), ())

    def test_equality_set(self):
        """Test for equality set map generation."""
        m = self.build_model()
        eq_var_map = _build_equality_set(m)
        self.assertIsNone(eq_var_map.get(m.z1, None))
        self.assertIsNone(eq_var_map.get(m.v1, None))
        self.assertIsNone(eq_var_map.get(m.v2, None))
        self.assertEqual(eq_var_map[m.v3], ComponentSet([m.v3, m.v4]))
        self.assertEqual(eq_var_map[m.v4], ComponentSet([m.v3, m.v4]))
        self.assertEqual(
            eq_var_map[m.x[1]], ComponentSet([m.x[1], m.x[2], m.x[3], m.x[4]])
        )
        self.assertEqual(
            eq_var_map[m.x[2]], ComponentSet([m.x[1], m.x[2], m.x[3], m.x[4]])
        )
        self.assertEqual(
            eq_var_map[m.x[3]], ComponentSet([m.x[1], m.x[2], m.x[3], m.x[4]])
        )
        self.assertEqual(
            eq_var_map[m.x[4]], ComponentSet([m.x[1], m.x[2], m.x[3], m.x[4]])
        )
        self.assertEqual(eq_var_map[m.y[1]], ComponentSet([m.y[1], m.y[2]]))
        self.assertEqual(eq_var_map[m.y[2]], ComponentSet([m.y[1], m.y[2]]))

    def test_var_aggregate(self):
        """Test for transitivity in a variable equality set."""
        m = self.build_model()

        TransformationFactory('contrib.aggregate_vars').apply_to(m)
        z_to_vars = m._var_aggregator_info.z_to_vars
        var_to_z = m._var_aggregator_info.var_to_z
        z = m._var_aggregator_info.z
        self.assertEqual(z_to_vars[z[1]], ComponentSet([m.v3, m.v4]))
        self.assertEqual(
            z_to_vars[z[2]], ComponentSet([m.x[1], m.x[2], m.x[3], m.x[4]])
        )
        self.assertEqual(z_to_vars[z[3]], ComponentSet([m.y[1], m.y[2]]))
        self.assertIs(var_to_z[m.v3], z[1])
        self.assertIs(var_to_z[m.v4], z[1])
        self.assertIs(var_to_z[m.x[1]], z[2])
        self.assertIs(var_to_z[m.x[2]], z[2])
        self.assertIs(var_to_z[m.x[3]], z[2])
        self.assertIs(var_to_z[m.x[4]], z[2])
        self.assertIs(var_to_z[m.y[1]], z[3])
        self.assertIs(var_to_z[m.y[2]], z[3])

        self.assertEqual(z[1].value, 2)
        self.assertEqual(z[1].lb, 2)
        self.assertEqual(z[1].ub, 4)

        self.assertEqual(z[3].value, 3.5)

    def test_min_if_not_None(self):
        self.assertEqual(min_if_not_None([1, 2, None, 3, None]), 1)
        self.assertEqual(min_if_not_None([None, None, None]), None)
        self.assertEqual(min_if_not_None([]), None)
        self.assertEqual(min_if_not_None([None, 3, -1, 2]), -1)
        self.assertEqual(min_if_not_None([0]), 0)
        self.assertEqual(min_if_not_None([0, None]), 0)

    def test_max_if_not_None(self):
        self.assertEqual(max_if_not_None([1, 2, None, 3, None]), 3)
        self.assertEqual(max_if_not_None([None, None, None]), None)
        self.assertEqual(max_if_not_None([]), None)
        self.assertEqual(max_if_not_None([None, 3, -1, 2]), 3)
        self.assertEqual(max_if_not_None([0]), 0)
        self.assertEqual(max_if_not_None([0, None]), 0)

    @unittest.skipIf(
        not SolverFactory('glpk').available(), "GLPK solver is not available."
    )
    def test_var_update(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var(bounds=(0, 1))
        m.c = Constraint(expr=m.x == m.y)
        m.o = Objective(expr=m.x)
        TransformationFactory('contrib.aggregate_vars').apply_to(m)
        SolverFactory('glpk').solve(m)
        z = m._var_aggregator_info.z
        self.assertEqual(z[1].value, 0)
        self.assertEqual(m.x.value, None)
        self.assertEqual(m.y.value, None)
        TransformationFactory('contrib.aggregate_vars').update_variables(m)
        self.assertEqual(z[1].value, 0)
        self.assertEqual(m.x.value, 0)
        self.assertEqual(m.y.value, 0)

    def test_binary_inequality(self):
        m = ConcreteModel()
        m.x = Var(domain=Binary)
        m.y = Var(domain=Binary)
        m.c = Constraint(expr=m.x == m.y)
        m.o = Objective(expr=0.5 * m.x + m.y, sense=maximize)
        TransformationFactory('contrib.aggregate_vars').apply_to(m)
        var_to_z = m._var_aggregator_info.var_to_z
        z = var_to_z[m.x]
        self.assertIs(var_to_z[m.y], z)
        self.assertEqual(z.domain, Binary)
        self.assertEqual(z.lb, 0)
        self.assertEqual(z.ub, 1)
        assertExpressionsEqual(self, m.o.expr, 0.5 * z + z)

    def test_equality_different_domains(self):
        m = ConcreteModel()
        m.x = Var(domain=Reals, bounds=(1, 2))
        m.y = Var(domain=Binary)
        m.c = Constraint(expr=m.x == m.y)
        m.o = Objective(expr=0.5 * m.x + m.y, sense=maximize)
        TransformationFactory('contrib.aggregate_vars').apply_to(m)
        var_to_z = m._var_aggregator_info.var_to_z
        z = var_to_z[m.x]
        self.assertIs(var_to_z[m.y], z)
        self.assertEqual(z.lb, 1)
        self.assertEqual(z.ub, 1)
        self.assertEqual(z.domain, Binary)
        assertExpressionsEqual(self, m.o.expr, 0.5 * z + z)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""This module provides standard :py:class:`enum.Enum` definitions used in
Pyomo, along with additional utilities for working with custom Enums

Utilities:

.. autosummary::

   ExtendedEnumType
   NamedIntEnum

Standard Enums:

.. autosummary::

   ObjectiveSense
   SolverAPIVersion

"""

import enum
import itertools
import re
import sys

# Aliases for enum members so that clients can use
# :py:mod:`pyomo.common.enums` like they would :py:mod:`enum`.
from enum import Enum, Flag

if sys.version_info[:2] < (3, 11):
    EnumType = enum.EnumMeta
else:
    EnumType = enum.EnumType
if sys.version_info[:2] < (3, 13):
    import inspect

    # prior to 3.13 the int.{to,from}_bytes docstrings had LaTeX-like
    # "`..'" quotations, which Sphinx can't parse correctly.
    def _fix_doc(ref):
        def _rewrite(func):
            func.__doc__ = re.sub(r"`(\S+)'", r"`\1`", ref.__doc__)
            return func

        return _rewrite

    class _int_bytes_doc_wrapper:
        """Class to wrap to_bytes/from_bytes and rewrite the docstring"""

        @_fix_doc(int.to_bytes)
        def to_bytes(self, /, length=1, byteorder='big', *, signed=False):
            return super().to_bytes(length=length, byteorder=byteorder, signed=signed)

        # Note: we need to use a decorator to set the __doc__ *before*
        # the @classmethod (which makes __doc__ read-only).
        @classmethod
        @_fix_doc(int.from_bytes)
        def from_bytes(cls, bytes, byteorder='big', *, signed=False):
            return super(_int_bytes_doc_wrapper, cls).from_bytes(
                bytes, byteorder=byteorder, signed=signed
            )

    def _doc_updater(base):
        return (
            inspect.cleandoc(
                f"""A compatibility wrapper around :class:`enum.{base.__name__}`

                This wrapper class updates the :meth:`to_bytes` and
                :meth:`from_bytes` docstrings in Python <= 3.12 to
                suppress warnings generated by Sphinx.

                .. rubric:: {base.__name__}

            """
            )
            + "\n\n"
            # There are environments where IntEnum.__doc__ is None (see #3710)
            + inspect.cleandoc(getattr(base, "__doc__", "") or "")
        )

    class IntEnum(_int_bytes_doc_wrapper, enum.IntEnum):
        __doc__ = _doc_updater(enum.IntEnum)

    class IntFlag(_int_bytes_doc_wrapper, enum.IntFlag):
        __doc__ = _doc_updater(enum.IntFlag)

else:
    IntEnum = enum.IntEnum
    IntFlag = enum.IntFlag


class ExtendedEnumType(EnumType):
    """Metaclass for creating an :py:class:`enum.Enum` that extends another Enum

    In general, :py:class:`enum.Enum` classes are not extensible: that is,
    they are frozen when defined and cannot be the base class of another
    Enum.  This Metaclass provides a workaround for creating a new Enum
    that extends an existing enum.  Members in the base Enum are all
    present as members on the extended enum.

    Example
    -------

    .. testcode::
       :hide:

       import enum
       from pyomo.common.enums import ExtendedEnumType

    .. testcode::

       class ObjectiveSense(enum.IntEnum):
           minimize = 1
           maximize = -1

       class ProblemSense(enum.IntEnum, metaclass=ExtendedEnumType):
           __base_enum__ = ObjectiveSense

           unknown = 0

    .. doctest::

       >>> list(ProblemSense)
       [<ProblemSense.unknown: 0>, <ObjectiveSense.minimize: 1>, <ObjectiveSense.maximize: -1>]
       >>> ProblemSense.unknown
       <ProblemSense.unknown: 0>
       >>> ProblemSense.maximize
       <ObjectiveSense.maximize: -1>
       >>> ProblemSense(0)
       <ProblemSense.unknown: 0>
       >>> ProblemSense(1)
       <ObjectiveSense.minimize: 1>
       >>> ProblemSense('unknown')
       <ProblemSense.unknown: 0>
       >>> ProblemSense('maximize')
       <ObjectiveSense.maximize: -1>
       >>> hasattr(ProblemSense, 'minimize')
       True
       >>> ProblemSense.minimize is ObjectiveSense.minimize
       True
       >>> ProblemSense.minimize in ProblemSense
       True

    """

    def __getattr__(cls, attr):
        try:
            return getattr(cls.__base_enum__, attr)
        except:
            return super().__getattr__(attr)

    def __iter__(cls):
        # The members of this Enum are the base enum members joined with
        # the local members
        return itertools.chain(super().__iter__(), cls.__base_enum__.__iter__())

    def __contains__(cls, member):
        # This enum "contains" both its local members and the members in
        # the __base_enum__ (necessary for good auto-enum[sphinx] docs)
        return super().__contains__(member) or member in cls.__base_enum__

    def __dir__(self):
        _dir = set(super().__dir__())
        _dir.update(e.name for e in self)
        return _dir

    def __instancecheck__(cls, instance):
        if cls.__subclasscheck__(type(instance)):
            return True
        # Also pretend that members of the extended enum are subclasses
        # of the __base_enum__.  This is needed to circumvent error
        # checking in enum.__new__ (e.g., for `ProblemSense('minimize')`)
        return cls.__base_enum__.__subclasscheck__(type(instance))

    def _missing_(cls, value):
        # Support attribute lookup by value or name
        for attr in ('value', 'name'):
            for member in cls:
                if getattr(member, attr) == value:
                    return member
        return None

    def __new__(metacls, cls, bases, classdict, **kwds):
        # Support lookup by name - but only if the new Enum doesn't
        # specify its own implementation of _missing_
        if '_missing_' not in classdict:
            classdict['_missing_'] = classmethod(ExtendedEnumType._missing_)
        return super().__new__(metacls, cls, bases, classdict, **kwds)


class NamedIntEnum(IntEnum):
    """An extended version of :py:class:`~pyomo.common.enums.IntEnum` that supports
    creating members by name as well as value.

    """

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.name == value:
                return member
        return None


class ObjectiveSense(NamedIntEnum):
    """Flag indicating if an objective is minimizing (1) or maximizing (-1).

    While the numeric values are arbitrary, there are parts of Pyomo
    that rely on this particular choice of value.  These values are also
    consistent with some solvers (notably Gurobi).

    """

    minimize = 1
    maximize = -1

    # Overloading __str__ is needed to match the behavior of the old
    # pyutilib.enum class (removed June 2020). There are spots in the
    # code base that expect the string representation for items in the
    # enum to not include the class name. New uses of enum shouldn't
    # need to do this.
    def __str__(self):
        return self.name


class SolverAPIVersion(NamedIntEnum):
    """
    Enum identifying Pyomo solver API version

    The numeric values are intentionally a bit odd because APPSI came
    between the official V1 and V2. We still want it to be chronologically
    in order without sacrificing the human-logic of v1 vs. v2.
    """

    #: Original Coopr/Pyomo solver interface
    V1 = 10
    #: Automatic Persistent Pyomo Solver Interface (experimental)
    APPSI = 15
    #: Redesigned solver interface (circa 2024)
    V2 = 20


minimize = ObjectiveSense.minimize
maximize = ObjectiveSense.maximize



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import weakref
from pyomo.common.collections import ComponentMap
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.set import UnknownSetDimen
from pyomo.core.base.var import Var
from pyomo.dae.contset import ContinuousSet


def create_access_function(var):
    """
    This method returns a function that returns a component by calling
    it rather than indexing it
    """

    def _fun(*args):
        return var[args]

    return _fun


class DAE_Error(Exception):
    """Exception raised while processing DAE Models"""


@ModelComponentFactory.register("Derivative of a Var in a DAE model.")
class DerivativeVar(Var):
    """
    Represents derivatives in a model and defines how a
    :py:class:`Var<pyomo.environ.Var>` is differentiated

    The :py:class:`DerivativeVar <pyomo.dae.DerivativeVar>` component is
    used to declare a derivative of a :py:class:`Var <pyomo.environ.Var>`.
    The constructor accepts a single positional argument which is the
    :py:class:`Var<pyomo.environ.Var>` that's being differentiated. A
    :py:class:`Var <pyomo.environ.Var>` may only be differentiated with
    respect to a :py:class:`ContinuousSet<pyomo.dae.ContinuousSet>` that it
    is indexed by. The indexing sets of a :py:class:`DerivativeVar
    <pyomo.dae.DerivativeVar>` are identical to those of the :py:class:`Var
    <pyomo.environ.Var>` it is differentiating.

    Parameters
    ----------
    sVar : ``pyomo.environ.Var``
        The variable being differentiated

    wrt : ``pyomo.dae.ContinuousSet`` or tuple
        Equivalent to `withrespectto` keyword argument. The
        :py:class:`ContinuousSet<pyomo.dae.ContinuousSet>` that the
        derivative is being taken with respect to. Higher order derivatives
        are represented by including the
        :py:class:`ContinuousSet<pyomo.dae.ContinuousSet>` multiple times in
        the tuple sent to this keyword. i.e. ``wrt=(m.t, m.t)`` would be the
        second order derivative with respect to ``m.t``
    """

    # Private Attributes:
    # _stateVar   The :class:`Var` being differentiated
    # _wrt        A list of the :class:`ContinuousSet` components the
    #             derivative is being taken with respect to
    # _expr       An expression representing the discretization equations
    #             linking the :class:`DerivativeVar` to its state :class:`Var`.

    def __init__(self, sVar, **kwds):
        if not isinstance(sVar, Var):
            raise DAE_Error(
                "%s is not a variable. Can only take the derivative of a Var"
                "component." % sVar
            )

        if "wrt" in kwds and "withrespectto" in kwds:
            raise TypeError(
                "Cannot specify both 'wrt' and 'withrespectto keywords "
                "in a DerivativeVar"
            )

        wrt = kwds.pop('wrt', None)
        wrt = kwds.pop('withrespectto', wrt)

        try:
            num_contset = len(sVar._contset)
        except AttributeError:
            # This dictionary keeps track of where the ContinuousSet appears
            # in the index. This implementation assumes that every element
            # in an indexing set has the same dimension.
            sVar._contset = ComponentMap()
            sVar._derivative = {}
            if sVar.dim() == 0:
                num_contset = 0
            else:
                sidx_sets = list(sVar.index_set().subsets())
                loc = 0
                for i, s in enumerate(sidx_sets):
                    if s.ctype is ContinuousSet:
                        sVar._contset[s] = loc
                    _dim = s.dimen
                    if _dim is None:
                        raise DAE_Error(
                            "The variable %s is indexed by a Set (%s) with a "
                            "non-fixed dimension.  A DerivativeVar may only be "
                            "indexed by Sets with constant dimension" % (sVar, s.name)
                        )
                    elif _dim is UnknownSetDimen:
                        raise DAE_Error(
                            "The variable %s is indexed by a Set (%s) with an "
                            "unknown dimension.  A DerivativeVar may only be "
                            "indexed by Sets with known constant dimension"
                            % (sVar, s.name)
                        )
                    loc += s.dimen
            num_contset = len(sVar._contset)

        if num_contset == 0:
            raise DAE_Error(
                "The variable %s is not indexed by any ContinuousSets. A "
                "derivative may only be taken with respect to a continuous "
                "domain" % sVar
            )

        if wrt is None:
            # Check to be sure Var is indexed by single ContinuousSet and take
            # first deriv wrt that set
            if num_contset != 1:
                raise DAE_Error(
                    "The variable %s is indexed by multiple ContinuousSets. "
                    "The desired ContinuousSet must be specified using the "
                    "keyword argument 'wrt'" % sVar
                )
            wrt = [next(iter(sVar._contset.keys()))]
        elif type(wrt) is ContinuousSet:
            if wrt not in sVar._contset:
                raise DAE_Error(
                    "Invalid derivative: The variable %s is not indexed by "
                    "the ContinuousSet %s" % (sVar, wrt)
                )
            wrt = [wrt]
        elif type(wrt) is tuple or type(wrt) is list:
            for i in wrt:
                if type(i) is not ContinuousSet:
                    raise DAE_Error(
                        "Cannot take the derivative with respect to %s. "
                        "Expected a ContinuousSet or a tuple of "
                        "ContinuousSets" % i
                    )
                if i not in sVar._contset:
                    raise DAE_Error(
                        "Invalid derivative: The variable %s is not indexed "
                        "by the ContinuousSet %s" % (sVar, i)
                    )
            wrt = list(wrt)
        else:
            raise DAE_Error(
                "Cannot take the derivative with respect to %s. "
                "Expected a ContinuousSet or a tuple of ContinuousSets" % i
            )

        wrtkey = [str(i) for i in wrt]
        wrtkey.sort()
        wrtkey = tuple(wrtkey)

        if wrtkey in sVar._derivative:
            raise DAE_Error(
                "Cannot create a new derivative variable for variable "
                "%s: derivative already defined as %s"
                % (sVar.name, sVar._derivative[wrtkey]().name)
            )

        sVar._derivative[wrtkey] = weakref.ref(self)
        self._sVar = sVar
        self._wrt = wrt

        kwds.setdefault('ctype', DerivativeVar)

        Var.__init__(self, sVar.index_set(), **kwds)

    def get_continuousset_list(self):
        """Return the a list of :py:class:`ContinuousSet` components the
        derivative is being taken with respect to.

        Returns
        -------
        `list`
        """
        return self._wrt

    def is_fully_discretized(self):
        """
        Check to see if all the
        :py:class:`ContinuousSets<pyomo.dae.ContinuousSet>` this derivative
        is taken with respect to have been discretized.

        Returns
        -------
        `boolean`
        """
        for i in self._wrt:
            if 'scheme' not in i.get_discretization_info():
                return False
        return True

    def get_state_var(self):
        """Return the :py:class:`Var` that is being differentiated.

        Returns
        -------
        :py:class:`Var<pyomo.environ.Var>`
        """
        return self._sVar

    def get_derivative_expression(self):
        """
        Returns the current discretization expression for this derivative or
        creates an access function to its :py:class:`Var` the first time
        this method is called. The expression gets built up as the
        discretization transformations are sequentially applied to each
        :py:class:`ContinuousSet` in the model.
        """
        try:
            return self._expr
        except:
            self._expr = create_access_function(self._sVar)
            return self._expr

    def set_derivative_expression(self, expr):
        """Sets``_expr``, an expression representing the discretization
        equations linking the :class:`DerivativeVar` to its state
        :class:`Var`
        """
        self._expr = expr



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger(__name__)

import pyomo.environ as pyo
from pyomo.common.collections import ComponentSet
from pyomo.contrib.alternative_solutions import Solution
import pyomo.contrib.alternative_solutions.aos_utils as aos_utils


def enumerate_binary_solutions(
    model,
    *,
    num_solutions=10,
    variables=None,
    rel_opt_gap=None,
    abs_opt_gap=None,
    search_mode="optimal",
    solver="gurobi",
    solver_options={},
    tee=False,
    seed=None,
):
    """
    Finds alternative optimal solutions for a binary problem using no-good
    cuts.

    This function implements a no-good cuts technique inheriting from
    Balas's work on Canonical Cuts [BJ72]_.

    Parameters
    ----------
    model : ConcreteModel
        A concrete Pyomo model
    num_solutions : int
        The maximum number of solutions to generate.
    variables: None or a collection of Pyomo _GeneralVarData variables
        The variables for which bounds will be generated. None indicates
        that all variables will be included. Alternatively, a collection of
        _GeneralVarData variables can be provided.
    rel_opt_gap : float or None
        The relative optimality gap for the original objective for which
        variable bounds will be found. None indicates that a relative gap
        constraint will not be added to the model.
    abs_opt_gap : float or None
        The absolute optimality gap for the original objective for which
        variable bounds will be found. None indicates that an absolute gap
        constraint will not be added to the model.
    search_mode : 'optimal', 'random', or 'hamming'
        Indicates the mode that is used to generate alternative solutions.
        The optimal mode finds the next best solution. The random mode
        finds an alternative solution in the direction of a random ray. The
        hamming mode iteratively finds solution that maximize the hamming
        distance from previously discovered solutions.
    solver : string
        The solver to be used.
    solver_options : dict
        Solver option-value pairs to be passed to the solver.
    tee : boolean
        Boolean indicating that the solver output should be displayed.
    seed : int
        Optional integer seed for the numpy random number generator

    Returns
    -------
    solutions
        A list of Solution objects.
        [Solution]

    """
    logger.info("STARTING NO-GOOD CUT ANALYSIS")

    assert search_mode in [
        "optimal",
        "random",
        "hamming",
    ], 'search mode must be "optimal", "random", or "hamming".'

    if seed is not None:
        aos_utils._set_numpy_rng(seed)

    all_variables = aos_utils.get_model_variables(model, include_fixed=True)
    if variables == None:
        binary_variables = [
            var for var in all_variables if var.is_binary() and not var.is_fixed()
        ]
        logger.debug(
            "Analysis using %d binary variables: %s"
            % (len(binary_variables), " ".join(var.name for var in binary_variables))
        )
    else:
        binary_variables = ComponentSet()
        non_binary_variables = []
        for var in variables:
            if var.is_binary():
                binary_variables.add(var)
            else:  # pragma: no cover
                non_binary_variables.append(var.name)
        if len(non_binary_variables) > 0:
            logger.warn(
                (
                    "Warning: The following non-binary variables were included"
                    "in the variable list and will be ignored:"
                )
            )
            logger.warn(", ".join(non_binary_variables))

    orig_objective = aos_utils.get_active_objective(model)

    if len(binary_variables) == 0:
        logger.warn("No binary variables found!")

    #
    # Setup solver
    #
    opt = pyo.SolverFactory(solver)
    opt.available()
    for parameter, value in solver_options.items():
        opt.options[parameter] = value
    #
    # Appsi-specific configurations
    #
    use_appsi = False
    if "appsi" in solver:
        use_appsi = True
        opt.update_config.update_constraints = False
        opt.update_config.check_for_new_or_removed_constraints = True
        opt.update_config.check_for_new_or_removed_vars = False
        opt.update_config.check_for_new_or_removed_params = False
        opt.update_config.update_vars = False
        opt.update_config.update_params = False
        opt.update_config.update_named_expressions = False
        opt.update_config.treat_fixed_vars_as_params = False

        if search_mode == "hamming":
            opt.update_config.check_for_new_objective = True
            opt.update_config.update_objective = True
        elif search_mode == "random":
            opt.update_config.check_for_new_objective = True
            opt.update_config.update_objective = False
        else:
            opt.update_config.check_for_new_objective = False
            opt.update_config.update_objective = False

    #
    # Initial solve of the model
    #
    logger.info("Performing initial solve of model.")
    results = opt.solve(model, tee=tee, load_solutions=False)
    status = results.solver.status
    if not pyo.check_optimal_termination(results):
        condition = results.solver.termination_condition
        raise Exception(
            (
                "No-good cut analysis cannot be applied, "
                "SolverStatus = {}, "
                "TerminationCondition = {}"
            ).format(status.value, condition.value)
        )

    model.solutions.load_from(results)
    orig_objective_value = pyo.value(orig_objective)
    logger.info("Found optimal solution, value = {}.".format(orig_objective_value))
    solutions = [Solution(model, all_variables, objective=orig_objective)]
    #
    # Return just this solution if there are no binary variables
    #
    if len(binary_variables) == 0:
        return solutions

    aos_block = aos_utils._add_aos_block(model, name="_balas")
    logger.info("Added block {} to the model.".format(aos_block))
    aos_block.no_good_cuts = pyo.ConstraintList()
    aos_utils._add_objective_constraint(
        aos_block, orig_objective, orig_objective_value, rel_opt_gap, abs_opt_gap
    )

    if search_mode in ["random", "hamming"]:
        orig_objective.deactivate()

    solution_number = 2
    while solution_number <= num_solutions:

        expr = 0
        for var in binary_variables:
            if var.value > 0.5:
                expr += 1 - var
            else:
                expr += var

        aos_block.no_good_cuts.add(expr=expr >= 1)

        if search_mode == "hamming":
            if hasattr(aos_block, "hamming_objective"):
                aos_block.hamming_objective.expr += expr
                if use_appsi and opt.update_config.check_for_new_objective:
                    opt.update_config.check_for_new_objective = False
            else:
                aos_block.hamming_objective = pyo.Objective(
                    expr=expr, sense=pyo.maximize
                )

        elif search_mode == "random":
            if hasattr(aos_block, "random_objective"):
                aos_block.del_component("random_objective")
            vector = aos_utils._get_random_direction(len(binary_variables))
            idx = 0
            expr = 0
            for var in binary_variables:
                expr += vector[idx] * var
                idx += 1
            aos_block.random_objective = pyo.Objective(expr=expr, sense=pyo.maximize)

        results = opt.solve(model, tee=tee, load_solutions=False)
        status = results.solver.status
        condition = results.solver.termination_condition
        if pyo.check_optimal_termination(results):
            model.solutions.load_from(results)
            orig_obj_value = pyo.value(orig_objective)
            logger.info(
                "Iteration {}: objective = {}".format(solution_number, orig_obj_value)
            )
            solutions.append(Solution(model, all_variables, objective=orig_objective))
            solution_number += 1
        elif (
            condition == pyo.TerminationCondition.infeasibleOrUnbounded
            or condition == pyo.TerminationCondition.infeasible
        ):
            logger.info(
                "Iteration {}: Infeasible, no additional binary solutions.".format(
                    solution_number
                )
            )
            break
        else:  # pragma: no cover
            logger.info(
                (
                    "Iteration {}: Unexpected condition, SolverStatus = {}, "
                    "TerminationCondition = {}"
                ).format(solution_number, status.value, condition.value)
            )
            break

    aos_block.deactivate()
    orig_objective.activate()

    logger.info("COMPLETED NO-GOOD CUT ANALYSIS")

    return solutions
